# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' glance_store/exceptions.py
GLANCE_STORE_EXCEPTIONS = [
    "BackendException",
    "UnsupportedBackend",
    "RedirectException",
    "GlanceStoreException",
    "MissingCredentialError",
    "BadAuthStrategy",
    "AuthorizationRedirect",
    "NotFound",
    "UnknownHashingAlgo",
    "UnknownScheme",
    "BadStoreUri",
    "Duplicate",
    "StorageFull",
    "StorageWriteDenied",
    "AuthBadRequest",
    "AuthUrlNotFound",
    "AuthorizationFailure",
    "NotAuthenticated",
    "Forbidden",
    "Invalid",
    "BadStoreConfiguration",
    "DriverLoadFailure",
    "StoreDeleteNotSupported",
    "StoreGetNotSupported",
    "StoreRandomGetNotSupported",
    "StoreAddDisabled",
    "MaxRedirectsExceeded",
    "NoServiceEndpoint",
    "RegionAmbiguity",
    "RemoteServiceUnavailable",
    "HasSnapshot",
    "InUseByStore",
    "HostNotInitialized",
]

# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' glance/common/exception.py
GLANCE_EXCEPTIONS = [
    "RedirectException",
    "GlanceException",
    "MissingCredentialError",
    "BadAuthStrategy",
    "NotFound",
    "BadStoreUri",
    "Duplicate",
    "Conflict",
    "StorageQuotaFull",
    "AuthBadRequest",
    "AuthUrlNotFound",
    "AuthorizationFailure",
    "NotAuthenticated",
    "UploadException",
    "Forbidden",
    "ForbiddenPublicImage",
    "ProtectedImageDelete",
    "ProtectedMetadefNamespaceDelete",
    "ProtectedMetadefNamespacePropDelete",
    "ProtectedMetadefObjectDelete",
    "ProtectedMetadefResourceTypeAssociationDelete",
    "ProtectedMetadefResourceTypeSystemDelete",
    "ProtectedMetadefTagDelete",
    "Invalid",
    "InvalidSortKey",
    "InvalidSortDir",
    "InvalidPropertyProtectionConfiguration",
    "InvalidSwiftStoreConfiguration",
    "InvalidFilterOperatorValue",
    "InvalidFilterRangeValue",
    "InvalidOptionValue",
    "ReadonlyProperty",
    "ReservedProperty",
    "AuthorizationRedirect",
    "ClientConnectionError",
    "ClientConfigurationError",
    "MultipleChoices",
    "LimitExceeded",
    "ServiceUnavailable",
    "ServerError",
    "UnexpectedStatus",
    "InvalidContentType",
    "BadRegistryConnectionConfiguration",
    "BadDriverConfiguration",
    "MaxRedirectsExceeded",
    "InvalidRedirect",
    "NoServiceEndpoint",
    "RegionAmbiguity",
    "WorkerCreationFailure",
    "SchemaLoadError",
    "InvalidObject",
    "ImageSizeLimitExceeded",
    "FailedToGetScrubberJobs",
    "ImageMemberLimitExceeded",
    "ImagePropertyLimitExceeded",
    "ImageTagLimitExceeded",
    "ImageLocationLimitExceeded",
    "SIGHUPInterrupt",
    "RPCError",
    "TaskException",
    "BadTaskConfiguration",
    "ImageNotFound",
    "TaskNotFound",
    "InvalidTaskStatus",
    "InvalidTaskType",
    "InvalidTaskStatusTransition",
    "ImportTaskError",
    "TaskAbortedError",
    "DuplicateLocation",
    "InvalidParameterValue",
    "InvalidImageStatusTransition",
    "MetadefDuplicateNamespace",
    "MetadefDuplicateObject",
    "MetadefDuplicateProperty",
    "MetadefDuplicateResourceType",
    "MetadefDuplicateResourceTypeAssociation",
    "MetadefDuplicateTag",
    "MetadefForbidden",
    "MetadefIntegrityError",
    "MetadefNamespaceNotFound",
    "MetadefObjectNotFound",
    "MetadefPropertyNotFound",
    "MetadefResourceTypeNotFound",
    "MetadefResourceTypeAssociationNotFound",
    "MetadefTagNotFound",
    "InvalidDataMigrationScript",
]



import abc
import os

import yaml
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log


class YDefsLoader():
    """ Load yaml definitions. """

    def __init__(self, ytype, filter_path=None):
        """
        @param ytype: the type of defs we are loading i.e. defs/<ytype>
        """
        self.ytype = ytype
        self._loaded_defs = None
        self.stats_num_files_loaded = 0
        self.filter_path = filter_path

    @staticmethod
    def _is_def(abs_path):
        return abs_path.endswith('.yaml')

    @staticmethod
    def _get_yname(path):
        return os.path.basename(path).partition('.yaml')[0]

    def _get_defs_recursive(self, path):
        """ Recursively find all yaml/files beneath a directory. """
        defs = {}
        for entry in os.listdir(path):
            abs_path = os.path.join(path, entry)
            if os.path.isdir(abs_path):
                subdefs = self._get_defs_recursive(abs_path)
                if subdefs:
                    defs[os.path.basename(abs_path)] = subdefs
            else:
                if not self._is_def(abs_path):
                    continue

                if self._get_yname(abs_path) == os.path.basename(path):
                    with open(abs_path, encoding='utf-8') as fd:
                        log.debug("applying dir globals %s", entry)
                        defs.update(yaml.safe_load(fd.read()) or {})

                    # NOTE: these files do not count towards the total loaded
                    # since they are only supposed to contain directory-level
                    # globals that apply to other definitions in or below this
                    # directory.
                    continue

                with open(abs_path, encoding='utf-8') as fd:
                    self.stats_num_files_loaded += 1
                    _content = yaml.safe_load(fd.read()) or {}
                    defs[self._get_yname(abs_path)] = _content

        return defs

    def _apply_filter(self, loaded):
        """
        If a path filter has been provided, exclude any/all properties that are
        not descendants of that path.
        """
        if not self.filter_path:
            return loaded

        groups = self.filter_path.split('.')
        for i, subgroup in enumerate(groups):
            if i == 0:
                loaded = {subgroup: loaded[subgroup]}
            else:
                prev = groups[i - 1]
                loaded[prev] = {subgroup: loaded[prev][subgroup]}

        return loaded

    @property
    def plugin_defs(self):
        """ Load yaml defs for the current plugin and type. """
        log.debug('loading %s definitions for plugin=%s', self.ytype,
                  HotSOSConfig.plugin_name)

        if self._loaded_defs:
            return self._loaded_defs

        path = os.path.join(HotSOSConfig.plugin_yaml_defs, self.ytype,
                            HotSOSConfig.plugin_name)
        # reset
        self.stats_num_files_loaded = 0
        if os.path.isdir(path):
            loaded = self._get_defs_recursive(path)
            log.debug("YDefsLoader: plugin %s loaded %s file(s)",
                      HotSOSConfig.plugin_name, self.stats_num_files_loaded)
            # only return if we loaded actual definitions (not just globals)
            if self.stats_num_files_loaded:
                loaded = self._apply_filter(loaded)
                self._loaded_defs = loaded
                return loaded

        return {}


class YHandlerBase():
    """
    Base class for all YAML handler types e.g. scenario and event handlers.
    """
    def __init__(self, global_searcher, *args, **kwargs):
        self.global_searcher = global_searcher
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def run(self):
        """ Process operations. """



# The following check property modules need to be imported so that the property
# classes within them get autoregistered. We do not add them to the interface
# list for this module since they are not accessed that way.
from hotsos.core.ycheck.engine.properties import (  # noqa: F403,F401
    checks,
    conclusions,
    inputdef,
    search,
    vardef,
)
from hotsos.core.ycheck.engine.properties.common import YDefsSection


__all__ = [
    YDefsSection.__name__,
    ]



import abc
import re

from hotsos.core.log import log
from hotsos.core.search import (
    SearchDef,
    SequenceSearchDef,
)
from hotsos.core.plugins.kernel.kernlog.common import (
    CallTraceHeuristicBase,
    CallTraceStateBase,
    TraceTypeBase,
    KernLogBase,
)

KERNLOG_TS = r'\[\s*\d+\.\d+\]'
KERNLOG_PREFIX = rf'(?:\S+\s+\d+\s+[\d:]+\s+\S+\s+\S+:\s+)?{KERNLOG_TS}'


class OOMTraceHeuristicCheckFreePages(CallTraceHeuristicBase):
    """
    Defines a basic memory check heuristic using information extracted from
    an OOM kill call trace.
    """
    def __init__(self, state):
        """
        @param state: an OOMCallTraceState object.
        """
        self.node = state.node_id
        self.zone = state.node_zone
        self.free = state.free
        self.min = state.min
        self.low = state.low

    def __call__(self):
        """
        Returns a list of strings representing failures.
        """
        fails = []
        if self.free < self.min:
            fails.append(f"Node {self.node} zone {self.zone} free pages "
                         f"{self.free} below min {self.min}")
        elif self.free < self.low:
            fails.append(f"Node {self.node} zone {self.zone} free pages "
                         f"{self.free} below low {self.low}")

        return fails


class OOMCallTraceState(CallTraceStateBase):
    """ Representation of OOM call trace state. """
    def __repr__(self):
        """
        Provide a nice way to view the state e.g. in debug logs.
        """
        prefix = ""
        if self.node_id:
            prefix = " -> "

        s = []
        for k, v in self._info.items():
            if k == 'nodes':
                s.append(f"{prefix}{k}:")
                for node, zones in v.items():
                    s.append(f">> node{node}")
                    for zone, fields in zones.items():
                        s.append(f" {zone}: {repr(fields)}")
            else:
                s.append(f"{prefix}{k}: {v}")

        # can't use formatted string with backslash < py312
        return "\n{}\n".format('\n'.join(s))  # noqa, pylint: disable=consider-using-f-string


class GenericTraceType(TraceTypeBase):
    """
    Used to identify call traces of any type.

    This is mainly useful for high level checks like whether or not any call
    traces have occurred.
    """

    def __init__(self):
        self._search_def = None
        self.generics = []

    @property
    def name(self):
        return 'calltrace-anytype'

    @property
    def searchdef(self):
        if self._search_def:
            return self._search_def

        # Very basic, just a simple expression to identify that a call trace
        # has occurred.
        expr = r'.+Call Trace:'
        self._search_def = SearchDef(expr, tag=self.name)
        return self._search_def

    def apply(self, results):
        """
        Run through the results.

        @param results: list of search.SearchResult objects.
        """
        log.debug("%s has %s results", self.__class__.__name__, len(results))
        for _trace in results:
            # just a save a value i.e. to make the list length represent the
            # number of call traces.
            self.generics.append(True)

    @property
    def heuristics(self):
        """
        Register any heuristics we want to run here.
        """
        return []

    def __len__(self):
        return len(self.generics)

    def __iter__(self):
        yield from self.generics


class MemFieldsBase():
    """
    Provides a common way to identify fields in an oom kill trace and extract
    their values.
    """

    def extract(self, part, line):
        for field in self.fields:
            ret = re.search(f'{field}:{self.expr}', line)
            if ret:
                part.add(field, ret.group(1))

    @property
    @abc.abstractmethod
    def expr(self):
        """ Search pattern template used to match a field and its value. """

    @property
    @abc.abstractmethod
    def fields(self):
        """ List of fields we want extract from a given section. """


class MemFieldsMain(MemFieldsBase):
    """
    Expression and fields for extracting process memory information from call
    traces.
    """
    @property
    def expr(self):
        return r'\s?(\d+)'

    @property
    def fields(self):
        return ['active_anon', 'inactive_anon', 'isolated_anon',
                'active_file', 'inactive_file', 'isolated_file', 'unevictable',
                'dirty', 'writeback', 'slab_reclaimable', 'slab_unreclaimable',
                'mapped', 'shmem', 'pagetables', 'bounce', 'free', 'free_pcp',
                'free_cma']


class MemFieldsNodeMem(MemFieldsBase):
    """
    Expression and fields for extracting process numa node-specific memory
    information from call traces.
    """
    @property
    def expr(self):
        return r'\s?(\d+)kB'

    @property
    def fields(self):
        # not all kernels have all these fields
        return ['active_anon', 'inactive_anon', 'active_file',
                'inactive_file', 'unevictable', r'isolated\(anon\)',
                r'isolated\(file\)', 'mapped', 'dirty', 'writeback', 'shmem',
                'shmem_thp', 'shmem_pmdmapped', 'anon_thp', 'writeback_tmp',
                'kernel_stack', 'pagetables', 'writeback_tmp', 'unstable']


class MemFieldsNodeUNRC(MemFieldsBase):
    """
    Expression and fields for extracting process unreclaimable memory
    information from call traces.
    """
    @property
    def expr(self):
        return r'\s?([a-z]+)$'

    @property
    def fields(self):
        return [r'all_unreclaimable\?']


class MemFieldsNodeZoneMem(MemFieldsBase):
    """
    Expression and fields for extracting process numa node-zone-specific memory
    information from call traces.
    """
    @property
    def expr(self):
        return r'\s?(\d+)[Kk]B'

    @property
    def fields(self):
        return ['free', 'min', 'low', 'high',
                'reserved_highatomic', 'active_anon', 'inactive_anon',
                'active_file', 'inactive_file', 'unevictable', 'writepending',
                'present', 'managed', 'mlocked', 'bounce', 'free_pcp',
                'local_pcp', 'free_cma']


class BcacheDeadlockType(TraceTypeBase):
    """
    Implementation of traceback analysis for bcache deadlocks.
    """
    def __init__(self):
        self._search_def = None
        self.deadlocks = []

    @property
    def name(self):
        return 'calltrace-bcache'

    @property
    def searchdef(self):
        if self._search_def:
            return self._search_def

        start = SearchDef(rf'{KERNLOG_PREFIX}.+\s+task bcache-register:(\d+) '
                          r'blocked for more than (\d+) seconds.')
        body = SearchDef(r'(\S+) schedule(\S+) closure_sync(\S+)'
                         r' bch_journal_meta(\S+) bch_btree_set_root(\S+)'
                         r' bch_journal_replay(\S+) register_bcache(\S+)')
        end = SearchDef(rf'{KERNLOG_PREFIX}.+\s+do_syscall_(\S+)')
        self._search_def = SequenceSearchDef(start, tag='bcache', body=body,
                                             end=end)
        return self._search_def

    def apply(self, results):
        """
        Run through the results.
        @param results: list of search.SearchResult objects.
        """
        log.debug("%s has %s results", self.__class__.__name__, len(results))
        if len(results) > 0:
            # we have found at least one deadlock
            self.deadlocks.append(True)

    @property
    def heuristics(self):
        """
        Register any heuristics we want to run here.
        """
        return []

    def __len__(self):
        return len(self.deadlocks)

    def __iter__(self):
        yield from self.deadlocks


class FanotifyDeadlockType(TraceTypeBase):
    """
    Implementation of traceback analysis for fanotify deadlocks.
    """
    def __init__(self):
        self._search_def = None
        self.fanotify_hangs = []

    @property
    def name(self):
        return 'calltrace-fanotify'

    @property
    def searchdef(self):
        if self._search_def:
            return self._search_def

        start = SearchDef(r'.+ blocked for more than (\d+) seconds.')
        body = SearchDef(r'(\S+) schedule(\S+) fanotify_handle_event(\S+)'
                         r' fsnotify(\S+)')
        end = SearchDef(r'.+\s+do_sys_open(\S+)')
        self._search_def = SequenceSearchDef(start, tag='fanotify', body=body,
                                             end=end)
        return self._search_def

    def apply(self, results):
        """
        Run through the results.
        @param results: list of search.SearchResult objects.
        """
        log.debug("%s has %s results", self.__class__.__name__, len(results))
        if len(results) > 0:
            # we have found at least one fanotify related hang
            self.fanotify_hangs.append(True)

    @property
    def heuristics(self):
        """
        Register any heuristics we want to run here.
        """
        return []

    def __len__(self):
        return len(self.fanotify_hangs)

    def __iter__(self):
        yield from self.fanotify_hangs


class OOMKillerTraceType(TraceTypeBase):
    """
    Implementation of traceback analysis for OOM tracebacks.
    """
    def __init__(self):
        self._search_def = None
        self.oom_traces = []

    @property
    def name(self):
        return 'oom-killer'

    @property
    def searchdef(self):
        if self._search_def:
            return self._search_def

        start = SearchDef(rf'{KERNLOG_PREFIX} (\S+) invoked oom-killer: .+, '
                          r'order=([+-]?\d+),.+')

        fields = set()
        for mfieldtype in [MemFieldsMain, MemFieldsNodeZoneMem,
                           MemFieldsNodeMem, MemFieldsNodeUNRC]:
            fields.update(mfieldtype().fields)

        # Only include lines that actually contain fields we want to extract
        body = SearchDef(rf"(.+(?:{'|'.join(fields)}).+)")
        # NOTE: we need a better way to capture the different variations of
        #       the following message.
        end = SearchDef(rf'{KERNLOG_PREFIX} '
                        r'(?:Out of memory: |Memory cgroup out of memory: )?'
                        r'Killed process (\d+) .+')
        self._search_def = SequenceSearchDef(start, tag='oom', body=body,
                                             end=end)
        return self._search_def

    def apply(self, results):
        log.debug("%s has %s results", self.__class__.__name__, len(results))
        for trace in results.values():
            oom_kill = OOMCallTraceState()
            oom_kill.add('nodes', {})
            current_node = None
            for item in trace:
                if item.tag.endswith('-start'):
                    oom_kill.add('procname', item.get(1))
                    oom_kill.add('order', item.get(2))
                elif item.tag.endswith('-body'):
                    # There will be one or more of these.
                    line = item.get(1)

                    ret = re.search(r"Node (\d+) active_anon:", line)
                    if ret:
                        current_node = OOMCallTraceState()
                        node_id = ret.group(1)
                        current_node.add('node_id', node_id)
                        oom_kill.nodes[node_id] = {'main': current_node}
                    else:
                        ret = re.search(r"Node (\d+) (\S+) free:", line)
                        if ret:
                            current_node = OOMCallTraceState()
                            node_id = ret.group(1)
                            node_zone = ret.group(2)
                            current_node.add('node_id', node_id)
                            current_node.add('node_zone', node_zone)
                            oom_kill.nodes[node_id][node_zone] = current_node

                    if not current_node:
                        MemFieldsMain().extract(oom_kill, line)
                    elif current_node.node_zone:
                        MemFieldsNodeZoneMem().extract(current_node, line)
                    else:
                        MemFieldsNodeMem().extract(current_node, line)
                        MemFieldsNodeUNRC().extract(current_node, line)
                else:
                    # end
                    oom_kill.add('pid', item.get(1))

            log.info(oom_kill)
            self.oom_traces.append(oom_kill)

    @property
    def heuristics(self):
        _heuristics = []
        for oom_trace in self.oom_traces:
            for node, zones in oom_trace.nodes.items():
                if node is None:
                    continue

                for zone, state in zones.items():
                    if zone == 'main':
                        continue

                    _heuristics.append(OOMTraceHeuristicCheckFreePages(state))

        return _heuristics

    def __len__(self):
        return len(self.oom_traces)

    def __iter__(self):
        yield from self.oom_traces


class HungtaskTraceType(TraceTypeBase):
    """
    Implementation of traceback analysis for hung task tracebacks.
    """
    def __init__(self):
        self._search_def = None
        self.hungtasks = []

    @property
    def name(self):
        return 'calltrace-hungtask'

    @property
    def searchdef(self):
        if self._search_def:
            return self._search_def

        start = SearchDef(r'.+ blocked for more than (\d+) seconds.')
        # We don't need the body so don't match/save it
        body = None
        end = SearchDef(r'.+\s+do_syscall_(\S+)')
        self._search_def = SequenceSearchDef(start, tag='hungtask', body=body,
                                             end=end)
        return self._search_def

    def apply(self, results):
        """
        Run through the results.
        @param results: list of searchtools.SearchResult objects.
        """
        log.debug("%s has %s results", self.__class__.__name__, len(results))
        if len(results) > 0:
            # we have found at least one blocked task
            self.hungtasks.append(True)

    @property
    def heuristics(self):
        """
        Register any heuristics we want to run here.
        """
        return []

    def __len__(self):
        return len(self.hungtasks)

    def __iter__(self):
        yield from self.hungtasks


class CallTraceManager(KernLogBase):
    """
    Manager for all call trace analysis types. From here all analysis is run
    and results collected.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # All trace types should be registered here.
        self.tracetypes = [GenericTraceType(), OOMKillerTraceType(),
                           BcacheDeadlockType(), HungtaskTraceType(),
                           FanotifyDeadlockType()]
        self.run()

    def run(self):
        for tracetype in self.tracetypes:
            self.searcher.add(tracetype.searchdef, self.path)

        self.results = self.searcher.run()
        for tracetype in self.tracetypes:
            if isinstance(tracetype.searchdef, SequenceSearchDef):
                results = self.results.find_sequence_sections(
                                                           tracetype.searchdef)
            else:
                results = self.results.find_by_tag(tracetype.searchdef.tag)

            tracetype.apply(results)

    def __getattr__(self, name):
        """
        If one or more trace has been identified with the given type name,
        return its object.
        """
        for h in self.tracetypes:
            if h.name == name.replace('_', '-'):
                return h

        return None



from hotsos.core.plugins.openstack.common import OpenstackBase, OpenStackChecks
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class OpenStackSummary(OpenstackBase, OpenStackChecks):
    """ Implementation of OpenStack summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    @staticmethod
    @summary_entry('neutron-l3ha', get_min_available_entry_index())
    def summary_neutron_l3ha():
        routers = {}
        ha_info = NeutronHAInfo()
        for router in ha_info.ha_routers:
            state = router.ha_state
            if state in routers:
                routers[state] += 1
            else:
                routers[state] = 1

        return routers or None




# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p'  ./heat/common/exception.py
HEAT_EXCEPTIONS = [
    "HeatException",
    "MissingCredentialError",
    "AuthorizationFailure",
    "NotAuthenticated",
    "Forbidden",
    "NotAuthorized",
    "Invalid",
    "UserParameterMissing",
    "UnknownUserParameter",
    "InvalidTemplateVersion",
    "InvalidTemplateSection",
    "ImmutableParameterModified",
    "InvalidMergeStrategyForParam",
    "ConflictingMergeStrategyForParam",
    "InvalidTemplateAttribute",
    "InvalidTemplateReference",
    "TemplateOutputError",
    "InvalidEncryptionKey",
    "InvalidExternalResourceDependency",
    "EntityNotFound",
    "PhysicalResourceExists",
    "PhysicalResourceNameAmbiguity",
    "PhysicalResourceIDAmbiguity",
    "InvalidTenant",
    "StackExists",
    "HeatExceptionWithPath",
    "StackValidationFailed",
    "InvalidSchemaError",
    "ResourceNotFound",
    "SnapshotNotFound",
    "InvalidGlobalResource",
    "ResourceTypeUnavailable",
    "InvalidBreakPointHook",
    "InvalidRestrictedAction",
    "ResourceNotAvailable",
    "ClientNotAvailable",
    "ResourceFailure",
    "NotSupported",
    "ResourceActionNotSupported",
    "ResourceActionRestricted",
    "ResourcePropertyConflict",
    "ResourcePropertyDependency",
    "ResourcePropertyValueDependency",
    "PropertyUnspecifiedError",
    "UpdateReplace",
    "ResourceUnknownStatus",
    "ResourceInError",
    "UpdateInProgress",
    "HTTPExceptionDisguise",
    "EgressRuleNotAllowed",
    "Error",
    "NotFound",
    "InvalidContentType",
    "RequestLimitExceeded",
    "DownloadLimitExceeded",
    "StackResourceLimitExceeded",
    "ActionInProgress",
    "ActionNotComplete",
    "StopActionFailed",
    "EventSendFailed",
    "InterfaceAttachFailed",
    "InterfaceDetachFailed",
    "UnsupportedObjectError",
    "OrphanedObjectError",
    "IncompatibleObjectVersion",
    "ObjectActionError",
    "ReadOnlyFieldError",
    "ConcurrentTransaction",
    "ObjectFieldInvalid",
    "KeystoneServiceNameConflict",
    "SIGHUPInterrupt",
    "InvalidServiceVersion",
    "InvalidTemplateVersions",
    "UnableToAutoAllocateNetwork",
]



from .common import CephChecks, CephConfig, CephDaemonAllOSDsFactory
from .cluster import CephCluster, CephCrushMap

__all__ = [
    CephChecks.__name__,
    CephConfig.__name__,
    CephDaemonAllOSDsFactory.__name__,
    CephCluster.__name__,
    CephCrushMap.__name__
    ]



import abc
import os
from dataclasses import dataclass, field
from typing import Any

import yaml
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log
from hotsos.core.utils import sorted_dict


class IssueContext():
    """ Key value store for machine readable context. """
    def __init__(self, **kwargs):
        self.context = {}
        self.set(**kwargs)

    def set(self, **kwargs):
        self.context.update(kwargs)

    def __len__(self):
        return len(self.context)

    def to_dict(self):
        return self.context


@dataclass(frozen=True)
class IssueEntry:
    """ A single issue object as stored for later use. """

    ref: str
    message: str
    key: str
    context: Any = None
    origin: str = field(
        # We're using default_factory here to defer the evaluation of
        # the "default" value to runtime.
        default_factory=(
            lambda: f"{HotSOSConfig.plugin_name}.{HotSOSConfig.part_name}"
        )
    )

    @property
    def content(self):
        _content = {self.key: self.ref,
                    'message': self.message,
                    'origin': self.origin}
        if HotSOSConfig.machine_readable:
            if len(self.context or {}):
                _content['context'] = self.context.to_dict()

        return _content


class IssuesStoreBase(abc.ABC):
    """ Base class for issue store backend implementatios. """
    def __init__(self):
        if not os.path.isdir(HotSOSConfig.plugin_tmp_dir):
            raise FileNotFoundError(
                f"plugin tmp dir '{HotSOSConfig.plugin_tmp_dir}' not found")

    @property
    @abc.abstractmethod
    def store_path(self):
        pass

    @abc.abstractmethod
    def load(self):
        pass

    @abc.abstractmethod
    def add(self, issue):
        pass


class KnownBugsStore(IssuesStoreBase):
    """ Known bug backend store """
    @property
    def store_path(self):
        return os.path.join(HotSOSConfig.plugin_tmp_dir, 'known_bugs.yaml')

    def load(self):
        """
        Fetch the current plugin known_bugs.yaml if it exists and return its
        contents or None if it doesn't exist yet.
        """
        if not os.path.exists(self.store_path):
            return {}

        with open(self.store_path, encoding='utf-8') as fd:
            bugs = yaml.safe_load(fd)
        if bugs and IssuesManager.SUMMARY_OUT_BUGS_ROOT in bugs:
            return bugs

        return {}

    def add(self, issue, context=None):
        #  def __init__(self, ref, message, key, context=None):
        entry = IssueEntry(
            ref=issue.url,
            message=issue.msg,
            key='id',
            context=context
        )
        current = self.load()
        if current:
            current[IssuesManager.SUMMARY_OUT_BUGS_ROOT].append(entry.content)
        else:
            current = {IssuesManager.SUMMARY_OUT_BUGS_ROOT: [entry.content]}

        with open(self.store_path, 'w', encoding='utf-8') as fd:
            fd.write(yaml.dump(current))


class IssuesStore(IssuesStoreBase):
    """ Potential issue backend store """
    @property
    def store_path(self):
        return os.path.join(HotSOSConfig.plugin_tmp_dir, 'issues.yaml')

    def load(self):
        """
        Fetch the current plugin issues.yaml if it exists and return its
        contents or None if it doesn't exist yet.
        """
        if not os.path.exists(self.store_path):
            return {}

        with open(self.store_path, encoding='utf-8') as fd:
            issues = yaml.safe_load(fd)
        if issues and IssuesManager.SUMMARY_OUT_ISSUES_ROOT in issues:
            return issues

        return {}

    def add(self, issue, context=None):
        """
        Fetch the current plugin issues.yaml if it exists and add new issue.
        """
        entry = IssueEntry(
            ref=issue.name,
            message=issue.msg,
            key='type',
            context=context
        )
        current = self.load()
        key = IssuesManager.SUMMARY_OUT_ISSUES_ROOT
        if current:
            current[key].append(entry.content)
        else:
            current = {key: [entry.content]}

        with open(self.store_path, 'w', encoding='utf-8') as fd:
            fd.write(yaml.dump(current))


class IssuesManager():
    """ Manager for all issue store backends.  """
    SUMMARY_OUT_ISSUES_ROOT = 'potential-issues'
    SUMMARY_OUT_BUGS_ROOT = 'bugs-detected'

    def __init__(self):
        self.bugstore = KnownBugsStore()
        self.issuestore = IssuesStore()

    def load_bugs(self):
        bugs = self.bugstore.load()
        if bugs and not HotSOSConfig.machine_readable:
            urls = {}
            for bug in bugs[self.SUMMARY_OUT_BUGS_ROOT]:
                urls[bug['id']] = bug['message']

            bugs = {self.SUMMARY_OUT_BUGS_ROOT: sorted_dict(urls)}

        return bugs

    def load_issues(self):
        issues = self.issuestore.load()
        if issues and not HotSOSConfig.machine_readable:
            types = {}
            for issue in issues[self.SUMMARY_OUT_ISSUES_ROOT]:
                # pluralise the type for display purposes
                issue_type = f"{issue['type']}s"
                if issue_type not in types:
                    types[issue_type] = []

                types[issue_type].append(issue['message'])

            # Sort them to ensure consistency in output
            for issue_type, issues in types.items():
                types[issue_type] = sorted(issues)

            issues = {self.SUMMARY_OUT_ISSUES_ROOT: sorted_dict(types)}

        return issues

    def add(self, issue, context=None):
        if issue.ISSUE_TYPE in ('bug', 'cve'):
            log.debug("issue is a %s", issue.ISSUE_TYPE)
            self.bugstore.add(issue, context=context)
        else:
            self.issuestore.add(issue, context=context)



from hotsos.core.plugins.maas import MAASChecks


class MAASSummary(MAASChecks):
    """ Implementation of MAAS summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    # No custom entries defined here yet so currently relying on the defaults.



import os
from datetime import datetime
from functools import cached_property
from dataclasses import dataclass, field

from hotsos.core.config import HotSOSConfig
from hotsos.core import host_helpers
from hotsos.core.log import log
from hotsos.core.plugins.openstack.exceptions import (
    EXCEPTIONS_COMMON,
)
from hotsos.core.plugins.openstack.exceptions_barbican import (
    BARBICAN_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_castellan import (
    CASTELLAN_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_cinder import (
    CINDER_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_designate import (
    DESIGNATE_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_glance import (
    GLANCE_EXCEPTIONS,
    GLANCE_STORE_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_heat import (
    HEAT_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_keystone import (
    KEYSTONE_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_manila import (
    MANILA_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_masakari import (
    MASAKARI_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_neutron import (
    NEUTRON_EXCEPTIONS,
    NEUTRONCLIENT_EXCEPTIONS,
    OVSDBAPP_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_nova import (
    NOVA_EXCEPTIONS,
    OS_VIF_EXCEPTIONS,
    PLACEMENT_EXCEPTIONS,
    PYTHON_LIBVIRT_EXCEPTIONS,
)
from hotsos.core.plugins.openstack.exceptions_octavia import (
    OCTAVIA_EXCEPTIONS,
)


# NOTE(tpsilva): when updating this, refer to the Charmed Openstack supported
# versions page: https://ubuntu.com/openstack/docs/supported-versions
OST_EOL_INFO = {
    'caracal': datetime(2029, 4, 30),
    'bobcat': datetime(2024, 7, 30),
    'antelope': datetime(2026, 4, 30),
    'zed': datetime(2024, 4, 30),
    'yoga': datetime(2027, 4, 30),
    'xena': datetime(2023, 4, 30),
    'wallaby': datetime(2024, 4, 30),
    'victoria': datetime(2022, 4, 30),
    'ussuri': datetime(2030, 4, 30),
    'train': datetime(2021, 2, 28),
    'stein': datetime(2022, 4, 30),
    'rocky': datetime(2020, 2, 29),
    'queens': datetime(2028, 4, 30),
    'mitaka': datetime(2024, 4, 30)
}

# NOTE(dosaboy): Find package versions at:
# https://openstack-ci-reports.ubuntu.com/reports/cloud-archive/index.html
OST_REL_INFO = {
    'aodh-common': {
        'caracal': '1:18.0.0',
        'bobcat': '1:17.0.0',
        'antelope': '1:16.0.0',
        'zed': '1:15.0.0',
        'yoga': '1:14.0.0',
        'xena': '1:13.0.0',
        'wallaby': '1:12.0.0',
        'victoria': '11.0.0',
        'ussuri': '10.0.0',
        'train': '9.0.0',
        'stein': '8.0.0',
        'rocky': '7.0.0',
        'queens': '6.0.0'},
    'barbican-common': {
        'caracal': '2:18.0.0',
        'bobcat': '2:17.0.0',
        'antelope': '2:16.0.0',
        'zed': '2:15.0.0',
        'yoga': '2:14.0.0',
        'xena': '1:13.0.0',
        'wallaby': '1:12.0.0',
        'victoria': '1:11.0.0',
        'ussuri': '1:10.0.0',
        'train': '1:9.0.0',
        'stein': '1:8.0.0',
        'rocky': '1:7.0.0',
        'queens': '1:6.0.0'},
    'cinder-common': {
        'caracal': '2:24.0.0',
        'bobcat': '2:23.0.0',
        'antelope': '2:22.0.0',
        'zed': '2:21.0.0',
        'yoga': '2:20.0.0',
        'xena': '2:19.0.0',
        'wallaby': '2:18.0.0',
        'victoria': '2:17.0.0',
        'ussuri': '2:16.0.0',
        'train': '2:15.0.0',
        'stein': '2:14.0.0',
        'rocky': '2:13.0.0',
        'queens': '2:12.0.0'},
    'designate-common': {
        'caracal': '1:18.0.0',
        'bobcat': '1:17.0.0',
        'antelope': '1:16.0.0',
        'zed': '1:15.0.0',
        'yoga': '1:14.0.0',
        'xena': '1:13.0.0',
        'wallaby': '1:12.0.0',
        'victoria': '1:11.0.0',
        'ussuri': '1:10.0.0',
        'train': '1:9.0.0',
        'stein': '1:8.0.0',
        'rocky': '1:7.0.0',
        'queens': '1:6.0.0'},
    'glance-common': {
        'caracal': '2:28.0.1',
        'bobcat': '2:27.0.0',
        'antelope': '2:26.0.0',
        'zed': '2:25.0.0',
        'yoga': '2:24.0.0',
        'xena': '2:23.0.0',
        'wallaby': '2:22.0.0',
        'victoria': '2:21.0.0',
        'ussuri': '2:20.0.0',
        'train': '2:19.0.0',
        'stein': '2:18.0.0',
        'rocky': '2:17.0.0',
        'queens': '2:16.0.0'},
    'heat-common': {
        'caracal': '1:22.0.0',
        'bobcat': '1:21.0.0',
        'antelope': '1:20.0.0',
        'zed': '1:19.0.0',
        'yoga': '1:18.0.0',
        'xena': '1:17.0.0',
        'wallaby': '1:16.0.0',
        'victoria': '1:15.0.0',
        'ussuri': '1:14.0.0',
        'train': '1:13.0.0',
        'stein': '1:12.0.0',
        'rocky': '1:11.0.0',
        'queens': '1:10.0.0'},
    'ironic-common': {
        'caracal': '1:24.1.1',
        'bobcat': '1:23.0.0',
        'antelope': '1:21.2',
        'zed': '1:20.2.0',
        'yoga': '1:19.0.0'},
    'keystone': {
        'caracal': '2:25.0.0',
        'bobcat': '2:24.0.0',
        'antelope': '2:23.0.0',
        'zed': '2:22.0.0',
        'yoga': '2:21.0.0',
        'xena': '2:20.0.0',
        'wallaby': '2:19.0.0',
        'victoria': '2:18.0.0',
        'ussuri': '2:17.0.0',
        'train': '2:16.0.0',
        'stein': '2:15.0.0',
        'rocky': '2:14.0.0',
        'queens': '2:13.0.0',
        'pike': '2:12.0.0',
        'ocata': '2:11.0.0'},
    'masakari-common': {
        'caracal': '17.0.0',
        'bobcat': '16.0.0',
        'antelope': '15.0.0',
        'zed': '14.0.0',
        'yoga': '13.0.0',
        'xena': '12.0.0',
        'wallaby': '11.0.0',
        'victoria': '10.0.0',
        'ussuri': '9.0.0',
        'train': '8.0.0',
        'stein': '7.0.0',
        'rocky': '6.0.0'},
    'neutron-common': {
        'caracal': '2:24.0.0',
        'bobcat': '2:23.0.0',
        'antelope': '2:22.0.0',
        'zed': '2:21.0.0',
        'yoga': '2:20.0.0',
        'xena': '2:19.0.0',
        'wallaby': '2:18.0.0',
        'victoria': '2:17.0.0',
        'ussuri': '2:16.0.0',
        'train': '2:15.0.0',
        'stein': '2:14.0.0',
        'rocky': '2:13.0.0',
        'queens': '2:12.0.0',
        'pike': '2:11.0.0',
        'ocata': '2:10.0.0',
        'newton': '2:9.0.0',
        'mitaka': '2:8.0.0',
        'liberty': '2:7.0.0',
        'kilo': '1:2015.1.0',
        'juno': '1:2014.2.0',
        'icehouse': '1:2014.1.0'},
    'nova-common': {
        'caracal': '3:29.0.1',
        'bobcat': '3:28.0.0',
        'antelope': '3:27.0.0',
        'zed': '3:26.0.0',
        'yoga': '3:25.0.0',
        'xena': '3:24.0.0',
        'wallaby': '3:23.0.0',
        'victoria': '2:22.0.0',
        'ussuri': '2:21.0.0',
        'train': '2:20.0.0',
        'stein': '2:19.0.0',
        'rocky': '2:18.0.0',
        'queens': '2:17.0.0',
        'pike': '2:16.0.0',
        'ocata': '2:15.0.0',
        'newton': '2:14.0.0',
        'mitaka': '2:13.0.0',
        'liberty': '2:12.0.0',
        'kilo': '1:2015.1.0',
        'juno': '1:2014.2.0',
        'icehouse': '1:2014.1.0'},
    'octavia-common': {
        'caracal': '1:14.0.0',
        'bobcat': '1:13.0.0',
        'antelope': '1:12.0.0',
        'zed': '1:11.0.0',
        'yoga': '1:10.0.0',
        'xena': '1:9.0.0',
        'wallaby': '1:8.0.0',
        'victoria': '7.0.0',
        'ussuri': '6.0.0',
        'train': '5.0.0',
        'stein': '4.0.0',
        'rocky': '3.0.0'},
    'openstack-dashboard-common': {
        'caracal': '4:24.0.0',
        'bobcat': '4:23.3.0',
        'antelope': '4:23.1.0',
        'zed': '4:22.2.0',
        'yoga': '4:22.0.0',
        'xena': '4:20.0.0',
        'wallaby': '4:19.0.0',
        'victoria': '4:18.0.0',
        'ussuri': '3:18.0.0',
        'train': '3:16.0.0',
        'stein': '3:15.0.0',
        'rocky': '3:14.0.0',
        'queens': '3:13.0.0'},
    'placement-common': {
        'caracal': '1:11.0.0',
        'bobcat': '1:10.0.0',
        'antelope': '1:9.0.0',
        'zed': '1:8.0.0',
        'yoga': '1:7.0.0',
        'xena': '1:6.0.0',
        'wallaby': '1:5.0.0',
        'victoria': '4.0.0',
        'ussuri': '3.0.0',
        'train': '2.0.0',
        'stein': '1.0.0'}
}

OST_EXCEPTIONS = {'barbican': BARBICAN_EXCEPTIONS + CASTELLAN_EXCEPTIONS,
                  'cinder': CINDER_EXCEPTIONS + CASTELLAN_EXCEPTIONS,
                  'designate': DESIGNATE_EXCEPTIONS,
                  'glance': GLANCE_EXCEPTIONS + GLANCE_STORE_EXCEPTIONS,
                  'heat': HEAT_EXCEPTIONS,
                  'keystone': KEYSTONE_EXCEPTIONS,
                  'manila': MANILA_EXCEPTIONS,
                  'masakari': MASAKARI_EXCEPTIONS,
                  'neutron': NEUTRON_EXCEPTIONS + OVSDBAPP_EXCEPTIONS,
                  'nova': NOVA_EXCEPTIONS + PYTHON_LIBVIRT_EXCEPTIONS +
                  NEUTRONCLIENT_EXCEPTIONS + OS_VIF_EXCEPTIONS,
                  'octavia': OCTAVIA_EXCEPTIONS,
                  'placement': PLACEMENT_EXCEPTIONS,
                  }


class OpenstackConfig(host_helpers.IniConfigBase):
    """ Openstack config interface. """
    def __getattr__(self, key):
        return self.get(key)


@dataclass(frozen=True)
class OSTProjectParameters:
    """Parameters for OSTProject class.

        @param name: name of this project
        @param config: dict of config files keyed by a label used to identify
                       them. All projects should have a config file labelled
                       'main'.
        @param apt_core_alt: optional list of apt packages (regex) that are
                             used by this project where the name of the project
                             is not the same as the name used for its packages.
        @param systemd_masked_services: optional list of services that are
               expected to be masked in systemd e.g. if they are actually being
               run by apache.
        @param systemd_deprecated_services: optional list of services that are
               deprecated. This can be helpful if e.g. you want to skip checks
               for resources related to these services.
        @param systemd_extra_services: optional list of systemd services that
               are used. This is useful e.g. if components are run under
               apache or if a package runs components using services whose name
               don't match the name of the project.
        @param log_path_overrides: specify log path for a given service that
                                   overrides the default format i.e.
                                   /var/log/<project>/<svc>.log.
    """
    name: str = None
    config: dict = None
    apt_core_alt: list = None
    systemd_masked_services: list = field(default_factory=lambda: [])
    systemd_deprecated_services: list = field(default_factory=lambda: [])
    systemd_extra_services: list = field(default_factory=lambda: [])
    log_path_overrides: dict = field(default_factory=lambda: {})


@dataclass
class OSTProjectAptHelperParams:
    """ APTPackageHelper package matching expression lists. """
    deps: list
    core: list


@dataclass
class OSTProjectSystemdHelperParams:
    """SystemdHelper service matching expression lists. """
    extra_services: list
    masked_services: list
    deprecated_services: list


class OSTProject():
    """
    Representation of an Openstack project/service.

    Provides service state, version, packages etc.
    """
    SVC_VALID_SUFFIX = r'[0-9a-zA-Z-_]*'
    PY_CLIENT_PREFIX = r"python3?-{}\S*"

    def __init__(self, params: OSTProjectParameters):
        """
        @param name: name of this project
        @param config: dict of config files keyed by a label used to identify
                       them. All projects should have a config file labelled
                       'main'.
        @param apt_core_alt: optional list of apt packages (regex) that are
                             used by this project where the name of the project
                             is not the same as the name used for its packages.
        @param systemd_masked_services: optional list of services that are
               expected to be masked in systemd e.g. if they are actually being
               run by apache.
        @param systemd_deprecated_services: optional list of services that are
               deprecated. This can be helpful if e.g. you want to skip checks
               for resources related to these services.
        @param log_path_overrides: specify log path for a given service that
                                   overrides the default format i.e.
                                   /var/log/<project>/<svc>.log.
        @param systemd_extra_services: optional list of systemd services that
               are used. This is useful e.g. if components are run under
               apache or if a package runs components using services whose name
               don't match the name of the project.
        """
        self.name = params.name

        self.apt_params = OSTProjectAptHelperParams([self.PY_CLIENT_PREFIX.
                                                     format(params.name)],
                                                    [params.name])
        if params.apt_core_alt:
            self.apt_params.core.extend(params.apt_core_alt)
            for c in params.apt_core_alt:
                self.apt_params.deps.append(self.PY_CLIENT_PREFIX.format(c))

        self.config = {}
        if params.config:
            for label, path in params.config.items():
                path = os.path.join(
                    HotSOSConfig.data_root,
                    "etc",
                    params.name,
                    path
                )
                self.config[label] = OpenstackConfig(path)

        self.systemd_params = OSTProjectSystemdHelperParams(
                                params.systemd_extra_services,
                                params.systemd_masked_services,
                                params.systemd_deprecated_services)

        self.logs_path = os.path.join("var/log", params.name)
        self.log_path_overrides = params.log_path_overrides
        self.exceptions = EXCEPTIONS_COMMON + OST_EXCEPTIONS.get(
            params.name,
            []
        )

    @cached_property
    def installed(self):
        """ Return True if the openstack service is installed. """
        return bool(host_helpers.APTPackageHelper(
                        core_pkgs=self.apt_params.core).core)

    @cached_property
    def services_expr(self):
        exprs = [f'{self.name}{self.SVC_VALID_SUFFIX}']
        if self.systemd_params.extra_services:
            exprs += self.systemd_params.extra_services
        return exprs

    @cached_property
    def services(self):
        exprs = self.services_expr
        info = host_helpers.SystemdHelper(service_exprs=exprs)
        if not info.services:
            log.debug("no systemd services found for '%s' - trying pebble",
                      self.name)
            info = host_helpers.PebbleHelper(service_exprs=exprs)

        return info.services

    def log_paths(self, include_deprecated_services=True):
        """
        Returns tuples of daemon name, log path for each agent/daemon.
        """
        proj_manage = f"{self.name}-manage"
        path = os.path.join('var/log', self.name, f"{proj_manage}.log")
        yield proj_manage, [path]
        for svc in self.services:
            if (not include_deprecated_services and
                    svc in self.systemd_params.deprecated_services):
                continue

            path = os.path.join('var/log', self.name,
                                f"{svc}.log")
            yield svc, self.log_path_overrides.get(svc, [path])


class OSTProjectCatalog():
    """ Catalog of Openstack projects tracked by this plugin. """
    # Services that are not actually openstack projects but are used by them
    OST_SERVICES_DEPS = ['dnsmasq',
                         r'(?:nfs-)?ganesha\S*',
                         'haproxy',
                         rf"keepalived{OSTProject.SVC_VALID_SUFFIX}",
                         rf"vault{OSTProject.SVC_VALID_SUFFIX}",
                         r'qemu-system-\S+',
                         'radvd',
                         ]

    # Set of packages that any project can depend on
    APT_DEPS_COMMON = ['conntrack',
                       'dnsmasq',
                       'haproxy',
                       'keepalived',
                       'libvirt-daemon',
                       'libvirt-bin',
                       r'nfs-ganesha\S*',
                       r'python3?-oslo[.-]',
                       'qemu-kvm',
                       'radvd',
                       ]

    def __init__(self):
        self._projects = {}
        self.add('aodh', config={'main': 'aodh.conf'},
                 systemd_masked_services=['aodh-api'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/aodh/aodh-api.log']})
        self.add('barbican', config={'main': 'barbican.conf'},
                 systemd_masked_services=['barbican-api'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/barbican/barbican-api.log']})
        self.add('ceilometer', config={'main': 'ceilometer.conf'},
                 systemd_masked_services=['ceilometer-api'])
        self.add('cinder', config={'main': 'cinder.conf'},
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/apache2/cinder_*.log']})
        self.add('designate', config={'main': 'designate.conf'},
                 systemd_extra_services=['apache2'])
        self.add('glance', config={'main': 'glance-api.conf'},
                 systemd_extra_services=['apache2'])
        self.add('gnocchi', config={'main': 'gnocchi.conf'},
                 systemd_masked_services=['gnocchi-api'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/gnocchi/gnocchi-api.log']})
        self.add('heat', config={'main': 'heat.conf'},
                 systemd_extra_services=['apache2'])
        self.add('horizon', apt_core_alt=['openstack-dashboard'],
                 systemd_extra_services=['apache2'])
        self.add('ironic', config={'main': 'ironic.conf'})
        self.add('keystone', config={'main': 'keystone.conf'},
                 systemd_masked_services=['keystone'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/keystone/keystone.log']})
        self.add('neutron',
                 config={'main': 'neutron.conf',
                         'openvswitch-agent':
                         'plugins/ml2/openvswitch_agent.ini',
                         'l3-agent': 'l3_agent.ini',
                         'dhcp-agent': 'dhcp_agent.ini',
                         'ovn': 'ovn.ini',
                         'ml2': 'plugins/ml2/ml2_conf.ini'},
                 systemd_masked_services=['nova-api-metadata'],
                 systemd_extra_services=['apache2'],
                 systemd_deprecated_services=['neutron-lbaas-agent',
                                              'neutron-lbaasv2-agent',
                                              'neutron-fwaas-agent'])
        self.add('nova', config={'main': 'nova.conf'},
                 # See LP bug 1957760 for reason why neutron-server is added.
                 systemd_masked_services=['nova-api-os-compute',
                                          'neutron-server'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/apache2/nova-*.log',
                                      'var/log/nova/nova-api-wsgi.log']})
        self.add('manila', config={'main': 'manila.conf'},
                 systemd_masked_services=['manila-api'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/manila/manila-api.log']})
        self.add('masakari', config={'main': 'masakari.conf'},
                 systemd_masked_services=['masakari'],
                 systemd_extra_services=['apache2', 'masakari-engine'],
                 log_path_overrides={'apache2':
                                     ['var/log/apache2/masakari_error.log']})
        self.add('octavia', config={'main': 'octavia.conf',
                                    'amphora': 'amphora-agent.conf'},
                 systemd_masked_services=['octavia-api'],
                 apt_core_alt=[r'amphora-\S+'],
                 systemd_extra_services=['apache2', 'amphora-agent'],
                 log_path_overrides={'apache2':
                                     ['var/log/octavia/octavia-api.log']})
        self.add('placement', config={'main': 'placement.conf'},
                 systemd_masked_services=['placement'],
                 systemd_extra_services=['apache2'],
                 log_path_overrides={'apache2':
                                     ['var/log/apache2/*error.log']})
        self.add('swift', config={'main': 'swift-proxy.conf',
                                  'proxy': 'swift-proxy.conf'},
                 systemd_extra_services=['apache2'])

    def __getitem__(self, name):
        return self._projects[name]

    def __getattr__(self, name):
        if name in self._projects:
            return self._projects[name]

        raise AttributeError(f"project {name} not found")

    @cached_property
    def all(self):
        return self._projects

    @cached_property
    def service_exprs(self):
        # Expressions used to match openstack systemd services for each project
        exprs = []
        for p in self.all.values():
            if p.installed:
                exprs.extend(p.services_expr)

        exprs.extend(self.OST_SERVICES_DEPS)
        return exprs

    @cached_property
    def default_masked_services(self):
        """
        Returns a list of services that are expected to be marked as masked in
        systemd.
        """
        masked = []
        for p in self.all.values():
            masked += p.systemd_params.masked_services

        return sorted(masked)

    def add(self, name, *args, **kwargs):
        params = OSTProjectParameters(name=name, *args, **kwargs)
        self._projects[name] = OSTProject(params)

    @cached_property
    def packages_core_exprs(self):
        core = set()
        for p in self.all.values():
            core.update(p.apt_params.core)

        return list(core)

    @cached_property
    def packages_dep_exprs(self):
        deps = set(self.APT_DEPS_COMMON)
        for p in self.all.values():
            deps.update(p.apt_params.deps)

        return list(deps)


class OSTServiceBase():
    """ Representation of an Openstack service. """
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nethelp = host_helpers.HostNetworkingHelper()
        self.project = OSTProjectCatalog()[name]

    @cached_property
    def installed(self):
        """ Return True if the openstack service is installed. """
        return self.project.installed



class ScenarioException(Exception):
    """ Generic exception used with Scenarios. """


class UnsupportedFormatError(Exception):
    """Raised when a value is not in the expected format."""


class InvalidFileFormatError(Exception):
    """Raised when a file's format is not in the expected format."""


class MismatchError(Exception):
    """Generic mismatch exception type."""


class NameAlreadyRegisteredError(Exception):
    """Raised when a given name is already registered in an entity."""


class InvalidPathError(Exception):
    """Raised when a path is invalid."""


class NameNotSetError(Exception):
    """Raised when the name for an entity is absent."""


class AlreadyLoadedError(Exception):
    """Raised when an entity is expected to be loaded only once, but the load
    operation is called again."""


class NotEnoughParametersError(Exception):
    """Raised when an operation did not get enough parameters to operate."""


class MissingRequiredParameterError(Exception):
    """Raised when an operation did not get a parameter required
     for the operation."""


class UnexpectedParameterError(Exception):
    """Raised when an operation did get an unexpected parameter."""


class NoCallbacksRegisteredError(Exception):
    """Raised when a callback-based entity does not have a callback
    registered."""


class NotYetInitializedError(Exception):
    """Raised when an entity is not initialized where it already must be."""


class PreconditionError(Exception):
    """Raised when an operation's precondition is not met."""


class ExpectationNotMetError(Exception):
    """Raised when an operation's expectation is not met."""



from dataclasses import dataclass, field

from hotsos.core import plugintools
from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    PebbleHelper,
    SystemdHelper,
)
from hotsos.core.plugins.rabbitmq.report import RabbitMQReport

RMQ_SERVICES_EXPRS = [
    r"beam.smp",
    r"epmd",
    r"rabbitmq-server",
]
RMQ_PACKAGES = [
    r"rabbitmq-server",
]


@dataclass
class RabbitMQInstallInfo(InstallInfoBase):
    """ RabbitMQ installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=RMQ_PACKAGES))
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(
                                            service_exprs=RMQ_SERVICES_EXPRS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(
                                            service_exprs=RMQ_SERVICES_EXPRS))


class RabbitMQChecks(plugintools.PluginPartBase):
    """ Rabbitmq checks. """
    plugin_name = 'rabbitmq'
    plugin_root_index = 7

    def __init__(self, *args, **kwargs):
        super().__init__()
        RabbitMQInstallInfo().mixin(self)

    @property
    def report(self):
        return RabbitMQReport()

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        if RabbitMQInstallInfo().apt.core:
            return True

        return False



from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeWithOpsBase,
)


class YRequirementTypeProperty(YRequirementTypeWithOpsBase):
    """
    Property requires type property. Provides support for defining checks on
    python properties.
    """
    override_keys = ['property']
    override_autoregister = True

    @property
    @intercept_exception
    def _result(self):
        if not isinstance(self.content, dict):
            path = self.content
        else:
            path = self.content['path']

        actual = self.get_property(path)
        result = self.apply_ops(self.ops, opinput=actual)
        log.debug('requirement check: property %s %s (result=%s)',
                  path, self.ops_to_str(self.ops), result)
        self.cache.set('property', path)
        self.cache.set('ops', self.ops_to_str(self.ops))
        self.cache.set('value_actual', actual)
        return result



import tempfile
from collections import namedtuple
from operator import attrgetter

from hotsos.core.config import HotSOSConfig


def sorted_dict(d, key=None, reverse=False):
    """
    Return dictionary sorted using key. If no key provided sorted by dict keys.
    """
    if key is None:
        return dict(sorted(d.items(), key=lambda e: e[0], reverse=reverse))

    return dict(sorted(d.items(), key=key, reverse=reverse))


def mktemp_dump(data, prefix=None):
    """Create a temporary file under the current plugin tmp directory and write
    data to the file.
    """
    if prefix is not None:
        ftmp = tempfile.mktemp(prefix=prefix, dir=HotSOSConfig.plugin_tmp_dir)
    else:
        ftmp = tempfile.mktemp(dir=HotSOSConfig.plugin_tmp_dir)

    with open(ftmp, 'w', encoding='utf-8') as fd:
        fd.write(data)

    return ftmp


def seconds_to_date(secs):
    days = int(secs / 86400)
    hours = int(secs / 3600 % 24)
    mins = int(secs / 60 % 60)
    secs = int(secs % 60)
    return f'{days}d:{hours}h:{mins}m:{secs}s'


def sample_set_regressions(samples, ascending=True):
    """
    Takes a sample set and determines the number of times the sequence
    regresses i.e. values go the opposite direction than is expected.

    The default is to expect values to be increasing but this can be reversed
    by setting ascending=False.
    """
    if ascending:
        prev_reset = min(samples)
    else:
        prev_reset = max(samples)

    prev = prev_reset
    repetitions = 0
    for s in samples:
        if ascending and s >= prev:
            prev = s
        elif not ascending and s <= prev:
            prev = s
        else:
            repetitions += 1
            prev = prev_reset

    return repetitions


def string_to_int_or_float(value):
    try:
        return int(value)
    except ValueError:
        return float(value)


def sort_suffixed_integers(values, reverse=False):
    """
    Given a list of values that contain numbers suffixed with a
    k/K/m/M/g/G/t/T/p/P, sort them taking into account their suffix.

    @param values: list of integer or string values
    @param reverse: sort order
    @return: sorted list of values (original type is maintained)
    """
    reals = []
    suffix_exps = {'p': 5, 't': 4, 'g': 3, 'm': 2, 'k': 1}
    if reverse:
        valid_suffixes = ('p', 't', 'g', 'm', 'k')
    else:
        valid_suffixes = ('k', 'm', 'g', 't', 'p')

    Entry = namedtuple('entry', ('raw', 'actual'))
    for item in values:
        if isinstance(item, str):
            if len(item) > 1:
                value = item[:-1]
                suffix = item[-1].lower()
                if suffix in valid_suffixes:
                    value = string_to_int_or_float(value)
                    value = value * 1024 ** suffix_exps[suffix]
                    reals.append(Entry(item, value))
                    continue

        value = string_to_int_or_float(item)
        reals.append(Entry(item, value))

    reals.sort(key=attrgetter('actual'), reverse=reverse)
    return [t.raw for t in reals]



from hotsos.core.plugins.system.system import SystemBase
from hotsos.core import plugintools


class SystemChecks(SystemBase, plugintools.PluginPartBase):
    """ System checks. """
    plugin_name = 'system'
    plugin_root_index = 1

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        # Always run
        return True



import abc
import os
from enum import IntEnum, auto
from dataclasses import dataclass

import yaml
from jinja2 import FileSystemLoader, Environment
from hotsos.core.config import HotSOSConfig
from hotsos.core.issues import IssuesManager
from hotsos.core.log import log
from hotsos.core.ycheck.engine.common import YHandlerBase
from hotsos.core.ycheck.scenarios import YScenarioChecker
from hotsos.core.ycheck.common import GlobalSearcher
from hotsos.core.ycheck.events import EventsSearchPreloader
from hotsos.core.ycheck.scenarios import ScenariosSearchPreloader
from hotsos.core.exceptions import NameNotSetError

PLUGINS = {}
PLUGIN_RUN_ORDER = []


class SummaryOffsetConflict(Exception):
    """ If more than one entry competes for the same offset in the summary
    it could give unexpected results so we raise this exception.
    """


class PluginRegistryMeta(type):
    """
    Each plugin that wants to add output to the finally summary must register
    itself by implementing one or more classes that have this class as the
    metaclass at some point in their resolution chain and:

        * class attribute 'plugin_root_index' which should be set once per
          plugin (usually in a high level base class) and defines the order in
          which plugins will be run.

        * class attribute 'summary_part_index' set to the integer index it
          should appear in the summary entry for that plugin. These indexes
          start at 0 and are unique to a plugin.

    NOTE: at the time of writing these registrations are triggered by importing
          the plugin in hotsos.plugin_extensions.__init__.py so be sure to add
          new plugins there.
    """

    def __init__(cls, _name, _mro, members):
        name = members.get('plugin_name')
        if name:
            PLUGINS[name] = []
            PLUGIN_RUN_ORDER.append((name, cls.plugin_root_index))

        for subcls in cls.__mro__:
            index_key = 'summary_part_index'
            if hasattr(subcls, index_key):
                index = subcls.summary_part_index
                if index is None:
                    continue

                existing = [e[index_key] for e in PLUGINS[cls.plugin_name]]
                if index in existing:
                    raise SummaryOffsetConflict(f"plugin {name} has index "
                                                f"conflict {index}")

                PLUGINS[cls.plugin_name].append({index_key: index,
                                                 'runner': subcls})


class HOTSOSDumper(yaml.Dumper):  # pylint: disable=too-many-ancestors
    """ Custom yaml dumper that preserves order. """
    def increase_indent(self, flow=False, indentless=False):
        return super().increase_indent(flow, False)

    def represent_dict_preserve_order(self, data):
        return self.represent_dict(data.items())


def get_plugins_sorted():
    """ Return list of plugin names in the order they are to be run. """
    return [e[0] for e in sorted(PLUGIN_RUN_ORDER, key=lambda e: e[1])]


def yaml_dump(data):
    """
    This is our version of yaml.dump but ensuring the format/style that we
    want.
    """
    HOTSOSDumper.add_representer(
        dict,
        HOTSOSDumper.represent_dict_preserve_order)
    return yaml.dump(data, Dumper=HOTSOSDumper,
                     default_flow_style=False).rstrip("\n")


class HTMLFormatter:
    """
    Format the summary as html.

    Ref: https://iamkate.com/code/tree-views/
    """

    def __init__(self, hostname, max_level=2):
        """
        @param max_level: The HTML will be collapsible up to max_level
        """
        self.hostname = hostname
        self.max_level = max_level

    @staticmethod
    def render(context, template):
        # jinja 2.10.x really needs this to be a str and e.g. not a PosixPath
        templates_dir = str(HotSOSConfig.templates_path)
        if not os.path.isdir(templates_dir):
            raise FileNotFoundError(
                f"jinja templates directory not found: '{templates_dir}'")

        env = Environment(loader=FileSystemLoader(templates_dir))
        template = env.get_template(template)
        return template.render(context)

    @property
    def header(self):
        return self.render({'hostname': self.hostname}, 'header.html')

    @property
    def footer(self):
        with open(os.path.join(HotSOSConfig.templates_path,
                               'footer.html'), encoding='utf-8') as fd:
            return fd.read()

    def _expand_list(self, data, level):
        context = {'list_elements': [], 'indent': '    '}
        for item in data:
            context['list_elements'].append(self._expand(item, level))

        return self.render(context, 'content_list.html')

    def _expand_dict(self, data, level):
        context = {'dict_elements': {},
                   'indent': '    ',
                   'collapsible': level < self.max_level}
        if level == 1:
            context['class'] = 'class="tree"'

        for key, value in data.items():
            context['dict_elements'][key] = self._expand(value, level + 1)

        return self.render(context, 'content_dict.html')

    def _expand(self, data, level):
        """Expand the data object.

        @param data: The data object. This can be a dict, list, or a flat type
                     such as int or str.
        @param level: The current header level
        @return: A string expansion formatted in HTML of the data object.
        """
        if isinstance(data, dict):
            return self._expand_dict(data, level)
        if isinstance(data, list):
            return self._expand_list(data, level)

        return data

    def dump(self, data):
        """Convert the data (dict) into an html document.

        @param dist: The data
        @return: the html document as a string.
        """
        content = self._expand(data, 1)
        return self.header + content + self.footer


class MarkdownFormatter:
    """
    Formatter to convert output to Markdown.
    """

    def _expand_dict(self, data, level):
        level_prefix = ''.join(['#' for i in range(level)])
        markdown_output = ''
        for key in data:
            markdown_output += f'\n{level_prefix} {key}\n'
            markdown_output += self._expand(data[key], level + 1)

        return markdown_output

    @staticmethod
    def _expand_list(data):
        markdown_output = '\n%s' % ''.join(f'- {item}\n' for item in data)

        return markdown_output

    def _expand(self, data, level):
        """Expand the data object.

        @param data: The data object. This can be a dict, list, or a flat type
                     such as int or str.
        @param level: The current header level
        @return: string expansion formatted in markdown of the data object.
        """
        if isinstance(data, dict):
            return self._expand_dict(data, level)
        if isinstance(data, list):
            return self._expand_list(data)

        return f'\n{data}\n'

    def dump(self, data):
        """Convert the data (dict) into a markdown document.

        @param data: dict:- The data
        @return: str: The markdown document
        """
        markdown = '# hotsos summary\n' + self._expand(data, 2)
        return markdown.rstrip('\n')


@dataclass
class SummaryEntry:
    """ SummaryEntry data type. """

    data: object
    index: int = 0


def summary_entry(name, index=0):
    """Decorate a class member function to indicate that it is a summary
    function. Summary functions are automatically discovered by the
    PluginPartBase class and will be part of the summary output."""

    def real_decorator(func):
        # Store the summary metadata information on the function itself.
        func.__summary_elem_name = name  # pylint: disable=protected-access
        func.__summary_elem_index = index  # pylint: disable=protected-access
        return func
    return real_decorator


def get_min_available_entry_index():
    """ Return first index available after default entries. """
    return DefaultSummaryEntryIndexes.AVAILABLE


class DefaultSummaryEntryIndexes(IntEnum):
    """ Indexes used for default summary entries. """
    VERSION = 0
    RELEASE = auto()
    SERVICES = auto()
    SNAPS = auto()
    DPKG = auto()
    DOCKER_IMAGES = auto()
    # The following index represents the first one that is available for
    # implementations to use and they implement their indexes as an offset from
    # this value so that if it is increased they are automatically increased.
    AVAILABLE = auto()


class ApplicationSummaryBase(metaclass=PluginRegistryMeta):
    """ Common structure for application plugin and summary output.

    Individual application plugins should implement this class and extend to
    include information specific to their application and the summary output
    they would like to display.

    A common set of summary entries are provided. These can be overriden with
    a different entry key by clobbering the entry index. A corresponding set of
    default attributes is also provided to support these entries and these
    default to None so produce no output by default.
    """

    def __init__(self, *args, **kwargs):
        self.apt = None  # APTPackageHelper
        self.snaps = None  # SnapPackageHelper
        self.docker = None  # DockerImageHelper
        self.pebble = None  # PebbleHelper
        self.systemd = None  # SystemdHelper
        super().__init__(*args, **kwargs)

    @property
    def version(self):
        """ Optional application version. """
        return None

    @property
    def release_name(self):
        """ Optional application release_name. """
        return None

    @property
    def days_to_eol(self):
        """ Optional application days_to_eol. """
        return None

    @property
    def bind_interfaces(self):
        """
        Optionally implement this method to return a dictionary of network
        interfaces used by this application.
        """

    @classmethod
    def default_summary_entries(cls):
        return [e for e in dir(ApplicationSummaryBase)
                if str(e).startswith('summary_')]

    @summary_entry('version', DefaultSummaryEntryIndexes.VERSION)
    def summary_version(self):
        return self.version

    @summary_entry('release', DefaultSummaryEntryIndexes.RELEASE)
    def summary_release(self):
        if not all([self.release_name is not None,
                    self.days_to_eol is not None]):
            return None

        return {'name': self.release_name,
                'days-to-eol': self.days_to_eol}

    @summary_entry('services', DefaultSummaryEntryIndexes.SERVICES)
    def summary_services(self):
        """Get string info for running services."""
        if self.systemd is not None and self.systemd.services:
            return self.systemd.summary

        if self.pebble is not None and self.pebble.services:
            return self.pebble.summary

        return None

    @summary_entry('snaps', DefaultSummaryEntryIndexes.SNAPS)
    def summary_snaps(self):
        if self.snaps is None:
            return None

        return self.snaps.all_formatted or None

    @summary_entry('dpkg', DefaultSummaryEntryIndexes.DPKG)
    def summary_dpkg(self):
        # require at least one core package to be installed to include
        # this in the report.
        if self.apt is not None and self.apt.core:
            return self.apt.all_formatted

        return None

    @summary_entry('docker-images', DefaultSummaryEntryIndexes.DOCKER_IMAGES)
    def summary_docker_images(self):
        if self.docker is None:
            return None

        # require at least one core image to be in-use to include
        # this in the report.
        if self.docker.core:
            return self.docker.all_formatted

        return None


class PartOutputManager():
    """
    Collects output of all parts from a plugin and saves them in YAML format.
    """

    def save(self, data, index):
        """
        Save part output yaml in temporary location. These are collected and
        aggregated at the end of the plugin run.
        """
        log.debug("saving entry: %s (index=%s)", list(data.keys())[0], index)
        HOTSOSDumper.add_representer(
            dict,
            HOTSOSDumper.represent_dict_preserve_order)
        out = yaml.dump(data, Dumper=HOTSOSDumper,
                        default_flow_style=False).rstrip("\n")

        part_path = os.path.join(HotSOSConfig.plugin_tmp_dir,
                                 (f"{HotSOSConfig.plugin_name}."
                                  f"{HotSOSConfig.part_name}.part.yaml"))

        # don't clobber
        if os.path.exists(part_path):
            newpath = part_path
            i = 0
            while os.path.exists(newpath):
                i += 1
                newpath = f"{part_path}.{i}"

            part_path = newpath

        with open(part_path, 'w', encoding='utf-8') as fd:
            fd.write(out)

        self.add_to_index(index, part_path)

    @property
    def indexes(self):
        path = os.path.join(HotSOSConfig.plugin_tmp_dir, "index.yaml")
        if os.path.exists(path):
            with open(path, encoding='utf-8') as fd:
                return yaml.safe_load(fd.read()) or {}

        return {}

    def add_to_index(self, index, part):
        indexes = self.indexes
        path = os.path.join(HotSOSConfig.plugin_tmp_dir, "index.yaml")
        with open(path, 'w', encoding='utf-8') as fd:
            if index in indexes:
                indexes[index].append(part)
            else:
                indexes[index] = [part]

            fd.write(yaml.dump(indexes))

    @staticmethod
    def meld_part_output(data, existing):
        """
        Don't allow root level keys to be clobbered, instead just
        update them. This assumes that part subkeys will be unique.
        """
        remove_keys = []
        for key in data:
            if key in existing:
                if isinstance(existing[key], dict):
                    existing[key].update(data[key])
                    remove_keys.append(key)

        if remove_keys:
            for key in remove_keys:
                del data[key]

        existing.update(data)

    def all(self):
        if not self.indexes:
            return {}

        parts = {}
        for index in sorted(self.indexes):
            for part in self.indexes[index]:
                with open(part, encoding='utf-8') as fd:
                    part_yaml = yaml.safe_load(fd)

                    # Don't allow root level keys to be clobbered, instead just
                    # update them. This assumes that part subkeys will be
                    # unique.
                    self.meld_part_output(part_yaml, parts)

        return {HotSOSConfig.plugin_name: parts}


class PluginPartBase(ApplicationSummaryBase):
    """ This is the base class used for all plugins.

    Provides a standard set of methods that plugins will need as well as the
    plumbing needed to add their output to the summary.
    """
    # max per part before overlap
    PLUGIN_PART_INDEX_MAX = 500
    # Must be set to the name of the plugin context in which we are
    # running and at runtime should match HotSOSConfig.plugin_name.
    plugin_name = None
    summary_part_index = None

    def __init__(self, *args, global_searcher=None, **kwargs):
        """
        @param global_searcher: optional ...
        """
        self.global_searcher = global_searcher
        plugin_tmp = HotSOSConfig.plugin_tmp_dir

        if self.plugin_name is None:
            raise NameNotSetError(
                f"{self.__class__.__name__}.plugin_name must be "
                "set to a value that represents the name of the plugin")

        if not plugin_tmp or not os.path.isdir(plugin_tmp):
            raise FileNotFoundError(
                "plugin plugin_tmp_dir not initialised - exiting")

        super().__init__(*args, **kwargs)

    @classmethod
    @abc.abstractmethod
    def is_runnable(cls):
        """
        Must be implemented by all plugins to define at runtime whether they
        should run.
        """

    @property
    def summary_subkey_include_default_entries(self):
        """
        By default the default/base class summary entries will NOT be included
        in the output of a summary subkey. To override this behaviour you must
        override this property to return True.
        """
        return False

    @property
    def summary_subkey(self):
        """
        This can be optionally implemented in order to have all output of the
        current part placed under the keyname provided by this property i.e.
        <plugin_name>: <subkey>: ...

        By default the default/base class summary entries will NOT be included
        in the output of a summary subkey. To override this behaviour you must
        override summary_subkey_include_default_entries to return True.
        """
        return None

    @property
    def summary(self):
        """
        This can be optionally implemented as a way of pushing all results into
        the summary root for the current plugin part.

        The alternative is to implement methods called __summary_<key> such
        that the returned value is placed under <key>.
        """
        return None

    def get_summary_entries(self):
        """Return a list of summary functions defined in the current class
        and all the derived classes.

        Summary functions are the functions decorated with @summary_entry
        decorator.
        """

        result = {}
        # Iterate through all attributes of BaseClass
        for name in dir(self.__class__):
            # Get the attribute from the current class (self.__class__)
            attr = getattr(self.__class__, name)
            # Check if it's a function and has the specified attribute
            if not (callable(attr) and hasattr(attr, '__summary_elem_name')):
                continue

            exclude_defaults = (self.summary_subkey is not None or
                                self.summary_part_index > 0)
            if (attr.__name__ in self.default_summary_entries() and
                    exclude_defaults):
                # Exclude base class (default) entries from subkey parts unless
                # explicitly requested. Also only add default entries to the
                # topmost part.
                if self.summary_subkey_include_default_entries:
                    if self.summary_subkey is None:
                        continue
                elif self.summary_part_index > 0:
                    continue

            name = getattr(attr, '__summary_elem_name')
            index = getattr(attr, '__summary_elem_index')
            log.debug("found summary entry %s (index=%s, part_index=%s, "
                      "subkey=%s)",
                      name, index, self.summary_part_index,
                      self.summary_subkey)
            result[name] = [attr.__name__, index]
        return result

    @property
    def output(self):
        _output = {}
        if self.summary:
            for key, data in self.summary.items():
                _output[key] = SummaryEntry(data=data)

        # Discover the summary functions
        implicit_summary_fns = self.get_summary_entries()

        if implicit_summary_fns:
            # Sort the functions by their index
            implicit_summary_fns = dict(
                sorted(implicit_summary_fns.items(),
                       key=lambda item: item[1][1])
            )
            for name, (attr_name, index) in implicit_summary_fns.items():
                attr = getattr(self, attr_name)
                result = attr()
                if not result:
                    continue
                _output[name] = SummaryEntry(data=result, index=index)

        return _output

    @property
    def raw_output(self):
        out = self.output
        if not out:
            return {}

        return {key: entry.data for key, entry in out.items()}


class PluginRunner():
    """
    Run a plugin. This will run all requests parts of a plugin including
    defaults, providing a global searcher for the runtime of the plugin.
    """
    def __init__(self, plugin):
        self.plugin = plugin
        self.parts = PLUGINS[plugin]
        self.failed_parts = []
        self.part_mgr = PartOutputManager()
        self.issues_mgr = IssuesManager()

    def _load_global_searcher(self, global_searcher):
        """ Load event and scenario searches into global searcher. """
        # Load searches into the GlobalSearcher
        for preloader in [EventsSearchPreloader, ScenariosSearchPreloader]:
            try:
                preloader(global_searcher).run()
            # We really do want to catch all here since we don't care why
            # it failed but don't want to fail hard if it does.
            except Exception as exc:  # pylint: disable=W0718
                name = preloader.__name__
                self.failed_parts.append(name)
                log.exception("search preloader '%s' raised exception: %s",
                              name, exc)

    def _run_always_parts(self, global_searcher):
        """
        Execute parts that run regardless of plugin context.

        The following are executed as part of each plugin run (but not last).

        @param global_searcher: GlobalSearcher object
        """
        always_run = {'auto_scenario_check': YScenarioChecker}
        for name, always_part in always_run.items():
            # update current env to reflect actual part being run
            HotSOSConfig.part_name = name
            try:
                always_part(global_searcher).run()
            # We really do want to catch all here since we don't care why
            # it failed but don't want to fail hard if it does.
            except Exception as exc:  # pylint: disable=W0718
                self.failed_parts.append(name)
                log.exception("part '%s' raised exception: %s", name, exc)

            # NOTE: we don't expect these parts to produce any output
            # for the summary so we wont check for it (they only raise
            # issues and bugs which are handled independently).

    def _save_issues_to_summary(self):
        """ Save any failures or bugs to the summary. """
        if self.failed_parts:
            # always put these at the top
            self.part_mgr.save({'failed-parts': self.failed_parts}, index=0)

        bugs = self.issues_mgr.load_bugs()
        raised_issues = self.issues_mgr.load_issues()
        summary_end_index = PluginPartBase.PLUGIN_PART_INDEX_MAX ** 2

        # Add detected known_bugs and raised issues to end summary.
        if bugs:
            self.part_mgr.save(bugs, index=summary_end_index)

        # Add raised issues to summary.
        if raised_issues:
            self.part_mgr.save(raised_issues, index=summary_end_index)

    def _plugin_is_runnable(self):
        """
        Return True if any part of the current plugin is runnable.
        """
        for part_info in self.parts:
            if part_info['runner'].is_runnable():
                return True

        return False

    def _run_plugin_parts(self, global_searcher):
        """ Execute parts for the current plugin context.

        @param global_searcher: GlobalSearcher object
        @return: dictionary summary of output.
        """
        for part_info in self.parts:
            # update current env to reflect actual part being run
            runner = part_info['runner']
            name = runner.__name__
            HotSOSConfig.part_name = name
            if issubclass(runner, YHandlerBase):
                inst = runner(global_searcher)
            else:
                inst = runner(global_searcher=global_searcher)

            # Only run plugin if it declares itself runnable.
            if not HotSOSConfig.force_mode and not inst.is_runnable():
                log.debug("%s.%s not runnable - skipping",
                          HotSOSConfig.plugin_name, name)
                continue

            log.debug("running %s.%s", HotSOSConfig.plugin_name, name)
            try:
                # NOTE: since all parts are expected to be implementations
                # of PluginPartBase we expect them to always define an
                # output property.
                for key, entry in inst.output.items():
                    out = {key: entry.data}
                    if inst.summary_subkey:
                        out = {inst.summary_subkey: out}

                    part_max = PluginPartBase.PLUGIN_PART_INDEX_MAX
                    part_index = part_info['summary_part_index']
                    index = (part_index * part_max) + entry.index
                    self.part_mgr.save(out, index=index)
            # We really do want to catch all here since we don't care why
            # it failed but don't want to fail hard if it does.
            except Exception as exc:  # pylint: disable=W0718
                self.failed_parts.append(name)
                log.exception("part '%s' raised exception: %s", name, exc)

        # Add issues to summary.
        self._save_issues_to_summary()

        return self.part_mgr.all()

    def run(self):
        """ Execute all plugin parts. """
        self.failed_parts = []
        if not self._plugin_is_runnable():
            log.info("plugin '%s' not runnable - skipping", self.plugin)
            return {}

        with GlobalSearcher() as global_searcher:
            self._load_global_searcher(global_searcher)

            # Run the searches so that results are ready to be consumed when
            # the parts and handlers are run.
            global_searcher.run()

            self._run_always_parts(global_searcher)

            return self._run_plugin_parts(global_searcher)



from functools import cached_property

from hotsos.core.host_helpers import (
    SnapPackageHelper,
    DPKGVersion
)
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
    PackageCheckItemsBase,
)


class SnapCheckItems(PackageCheckItemsBase):
    """ Contains set of checks to run on snap check items. """
    @cached_property
    def packaging_helper(self):
        return SnapPackageHelper(core_snaps=self.packages_to_check)

    @cached_property
    def installed_revisions(self):
        return [self.packaging_helper.get_revision(p) for p in self.installed]

    @cached_property
    def installed_versions(self):
        return [self.packaging_helper.get_version(p) for p in self.installed]

    @cached_property
    def installed_channels(self):
        return [self.packaging_helper.get_channel(p) for p in self.installed]

    def package_info_matches(self, pkg, pkg_infos):
        """
        If snap package has revisions and/or channels we check them here.

        @param pkg: name of package
        @param pkg_info: list of {min: x, max: y, channel: z}
        """
        log.debug("checking snap package %s against %s",
                  str(pkg), str(pkg_infos))
        valid_keys = {"revision", "version", "channel"}

        for pkg_info in pkg_infos:
            # Check if there are unrecognized keys.
            if not pkg_info.keys() <= valid_keys:
                raise KeyError(
                    f"Unrecognized key name {pkg_info.keys()}."
                    f"Valid key names are {valid_keys}")

            if "revision" in pkg_info:
                sub = pkg_info["revision"]
                log.debug("revision criteria: %s", str(sub))
                # We're leveraging DPKG's version comparison algorithm
                # for the revisions as well.
                revision = self.packaging_helper.get_revision(pkg)
                result = DPKGVersion.is_version_within_ranges(revision, [sub])
                if not result:
                    continue
                log.debug("revision %s satisifies %s", str(revision), str(sub))

            if "version" in pkg_info:
                sub = pkg_info["version"]
                log.debug("version criteria: %s", str(sub))
                version = self.packaging_helper.get_version(pkg)
                result = DPKGVersion.is_version_within_ranges(version, [sub])
                if not result:
                    continue
                log.debug("version %s satisifies %s", str(version), str(sub))

            if "channel" in pkg_info:
                channel_v = pkg_info["channel"]
                log.debug("channel criteria: %s", str(channel_v))
                pkg_channel = self.packaging_helper.get_channel(pkg)
                if channel_v != pkg_channel:
                    continue
                log.debug("channel %s satisifies %s", pkg_channel, channel_v)

            # Package satisfies all criteria.
            return True
        # All checks failed.
        return False


class YRequirementTypeSnap(YRequirementTypeBase):
    """
    Snap requires type property. Provides support for defining checks on
    Ubuntu snaps.
    """
    override_keys = ['snap']
    override_autoregister = True

    @property
    def channel(self):
        return self.content.get('channel')

    @property
    @intercept_exception
    def _result(self):
        # pylint: disable=duplicate-code
        _result = True
        items = SnapCheckItems(self.content)
        # bail on first fail i.e. if any not installed
        if not items.not_installed:
            for pkg, info in items:
                if not info:
                    continue

                _result = items.package_info_matches(pkg, info)
                # bail at first failure
                if not _result:
                    break
        else:
            log.debug("one or more packages not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            _result = False

        log.debug('requirement check: snap(s) %s (result=%s)',
                  ', '.join(items.installed), _result)
        self.cache.set('package', ', '.join(items.installed))
        self.cache.set('revision', ', '.join(items.installed_revisions))
        self.cache.set('channel', ', '.join(items.installed_channels))
        self.cache.set('version', ', '.join(items.installed_versions))
        return _result



from hotsos.core.plugins.storage.ceph.common import CephChecks
from hotsos.core.utils import sorted_dict
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class CephSummary(CephChecks):
    """ Implementation of Ceph summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    @summary_entry('status', get_min_available_entry_index())
    def summary_status(self):
        return self.cluster.health_status or None

    @summary_entry('network', get_min_available_entry_index() + 1)
    def summary_network(self):
        """ Identify ports used by Ceph daemons, include them in output
        for informational purposes.
        """
        net_info = {}
        for config, port in self.bind_interfaces.items():
            net_info[config] = port.to_dict()

        if net_info:
            return net_info

        return None

    @summary_entry('osd-pgs-near-limit', get_min_available_entry_index() + 2)
    def summary_osd_pgs_near_limit(self):
        if self.cluster.osds_pgs_above_max:
            return self.cluster.osds_pgs_above_max

        return None

    @summary_entry('osd-pgs-suboptimal', get_min_available_entry_index() + 3)
    def summary_osd_pgs_suboptimal(self):
        if self.cluster.osds_pgs_suboptimal:
            return self.cluster.osds_pgs_suboptimal

        return None

    @summary_entry('versions',  get_min_available_entry_index() + 4)
    def summary_versions(self):
        versions = self.cluster.ceph_daemon_versions_unique()
        return versions or None

    @summary_entry('mgr-modules',  get_min_available_entry_index() + 5)
    def summary_mgr_modules(self):
        return self.cluster.mgr_modules or None

    @summary_entry('local-osds', get_min_available_entry_index() + 6)
    def summary_local_osds(self):
        if self.local_osds:
            osds = {}
            for osd in self.local_osds:
                osds.update(osd.to_dict())

            return sorted_dict(osds)

        return None

    @summary_entry('crush-rules', get_min_available_entry_index() + 7)
    def summary_crush_rules(self):
        return self.cluster.crush_map.rules or None

    @summary_entry('large-omap-pgs', get_min_available_entry_index() + 8)
    def summary_large_omap_pgs(self):
        return self.cluster.large_omap_pgs or None



import abc
from dataclasses import dataclass, field

from hotsos.core import plugintools
from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    PebbleHelper,
    SystemdHelper,
)
from hotsos.core.ycheck.events import EventCallbackBase, EventHandlerBase
from hotsos.core.ycheck.common import GlobalSearcherAutoRegisterBase
from hotsos.core.utils import sorted_dict

OVS_SERVICES_EXPRS = [r'ovsdb[a-zA-Z-]*',
                      r'ovs-vswitch[a-zA-Z-]*',
                      r'ovn[a-zA-Z-]*',
                      'openvswitch-switch',
                      ]
PY_CLIENT_PREFIX = r"python3?-{}\S*"
_OVS_PKGS_CORE = ['openvswitch-switch',
                  'ovn',
                  ]
OVS_PKGS_CORE = _OVS_PKGS_CORE + \
                [PY_CLIENT_PREFIX.format(p) for p in _OVS_PKGS_CORE]
_OVS_PKGS_DEPS = ['libc-bin',
                  'openvswitch',
                  'ovsdbapp',
                  'openssl',
                  'openvswitch-switch-dpdk',
                  ]
OVS_PKGS_DEPS = _OVS_PKGS_DEPS + \
                [PY_CLIENT_PREFIX.format(p) for p in _OVS_PKGS_DEPS]


class OpenvSwitchGlobalSearchBase(GlobalSearcherAutoRegisterBase):
    """ Base class for global searcher registration for OpenvSwitch. """
    plugin_name = "openvswitch"

    @classmethod
    @abc.abstractmethod
    def paths(cls):
        """ Returns a list of one or more paths to search. """


@dataclass
class OpenvSwitchInstallInfo(InstallInfoBase):
    """ OpenvSwitch installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=OVS_PKGS_CORE,
                                                   other_pkgs=OVS_PKGS_DEPS))
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(
                                            service_exprs=OVS_SERVICES_EXPRS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(
                                            service_exprs=OVS_SERVICES_EXPRS))


class OpenvSwitchChecks(plugintools.PluginPartBase):
    """ OpenvSwitch checks. """
    plugin_name = "openvswitch"
    plugin_root_index = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        OpenvSwitchInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        # require at least one core package to be installed to run this plugin.
        return len(OpenvSwitchInstallInfo().apt.core) > 0


class OpenvSwitchEventCallbackBase(EventCallbackBase):
    """ Base class for OpenvSwitch events callbacks. """
    @classmethod
    def global_event_tally_time_granularity_override(cls):
        return True

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """


class OpenvSwitchEventHandlerBase(OpenvSwitchChecks, EventHandlerBase):
    """ Base class for OpenvSwitch event handlers. """
    @property
    def summary(self):
        # mainline all results into summary root
        ret = self.run()
        if ret:
            return sorted_dict(ret)

        return None



import re
from functools import cached_property

from hotsos.core.log import log
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.host_helpers.common import ServiceManagerBase


class PebbleService():
    """ Representation of a pebble service. """
    def __init__(self, name, state):
        self.name = name
        self.state = state

    def __repr__(self):
        return f"name={self.name}, state={self.state}"


class PebbleHelper(ServiceManagerBase):
    """ Helper class used to query pebble services. """

    @property
    def _service_manager_type(self):
        return 'pebble'

    @cached_property
    def services(self):
        """ Return a dict of identified pebble services and their state. """
        _services = {}
        for line in CLIHelper().pebble_services():
            for svc_name_expr in self._service_exprs:
                _expr = rf"({svc_name_expr})\s+\S+\s+(\S+)\s+.*"
                ret = re.compile(_expr).match(line)
                if not ret:
                    continue

                name = ret.group(1)
                _services[name] = PebbleService(name=name, state=ret.group(2))

        return _services

    @cached_property
    def _service_filtered_ps(self):
        return CLIHelper().ps()


class ServiceFactory(FactoryBase):
    """
    Factory to dynamically create PebbleService objects for given services.

    Service objects are returned when a getattr() is done on this object using
    the name of the service as the attr name.
    """

    def __getattr__(self, svc):
        log.debug("creating service object for %s", svc)
        return PebbleHelper([svc]).services.get(svc)



from .common import OpenvSwitchChecks
from .ovs import (
    OpenvSwitchBase,
    OVSBFD,
    OVSDB,
    OVSDPLookups,
    OVSBridge,
    OVSDPDK,
)

__all__ = [
    OpenvSwitchChecks.__name__,
    OpenvSwitchBase.__name__,
    OVSBFD.__name__,
    OVSBridge.__name__,
    OVSDB.__name__,
    OVSDPDK.__name__,
    OVSDPLookups.__name__,
    ]



import re

from hotsos.core.host_helpers import CLIHelper, UptimeHelper
from hotsos.core.plugins.system import SystemChecks
from hotsos.core.plugintools import summary_entry


class SystemSummary(SystemChecks):
    """ Implementation of System summary. """
    summary_part_index = 0

    @summary_entry('hostname', 0)
    def summary_hostname(self):
        return self.hostname

    @summary_entry('os', 1)
    def summary_os(self):
        return self.os_release_name or None

    @summary_entry('num-cpus', 2)
    def summary_num_cpus(self):
        return self.num_cpus or None

    @staticmethod
    @summary_entry('load', 3)
    def summary_load():
        return UptimeHelper().loadavg or None

    @summary_entry('virtualisation', 4)
    def summary_virtualisation(self):
        return self.virtualisation_type or None

    @staticmethod
    @summary_entry('rootfs', 5)
    def summary_rootfs():
        df_output = CLIHelper().df()
        if df_output:
            for line in df_output:
                ret = re.compile(r"(.+\/$)").match(line)
                if ret:
                    return ret[1]

        return None

    @summary_entry('unattended-upgrades', 6)
    def summary_unattended_upgrades(self):
        if self.unattended_upgrades_enabled:
            return "ENABLED"
        return "disabled"

    @summary_entry('date', 7)
    def summary_date(self):
        return self.date

    @summary_entry('ubuntu-pro', 8)
    def summary_ubuntu_pro(self):
        return self.ubuntu_pro_status

    @summary_entry('uptime', 9)
    def summary_uptime(self):
        return self.uptime



import os
import re

from hotsos.core.config import HotSOSConfig
from hotsos.core.utils import sorted_dict


class _BaseProcKeyValue():
    """
    Base class for reading out of /proc/... files that have row/column
    structure that can be extracted into key/value format.
    """
    # Set to list of keys we expect to find in the file. If these keys are not
    # found their value will be returned as DEFAULT_RETURN_VALUE instead of
    # raising an AttributeError.
    VALID_KEYS = []
    DEFAULT_RETURN_VALUE = 0

    def __init__(self):
        self._proc_file_keys_values = {}

    @property
    def path(self):
        raise NotImplementedError('Method "path" must be implemened in '
                                  f'{self.__class__.__name__}.')

    @staticmethod
    def key_val_sep():
        """
        Returns regexp pattern for separating key and value in the
        /proc/<file>. Child class can override the default if the <file>
        structure is more complex.
        """
        return ' '

    def __getattr__(self, key):
        if key in self._proc_file_keys_values:
            return self._proc_file_keys_values[key]

        if not os.path.exists(self.path):
            return 0

        with open(self.path, encoding='utf-8') as fd:
            for line in fd:
                if re.split(self.key_val_sep(), line, 1)[0] == key:
                    value = int(re.split(self.key_val_sep(), line)[1])
                    self._proc_file_keys_values[key] = value
                    return value

        if key in self.VALID_KEYS:
            return self.DEFAULT_RETURN_VALUE

        raise AttributeError(f'attribute {key} not found in '
                             f'{self.__class__.__name__}.')


class VMStat(_BaseProcKeyValue):
    """ Interface to /proc/vmstat """
    VALID_KEYS = ['compact_fail', 'compact_success']

    @property
    def path(self):
        return os.path.join(HotSOSConfig.data_root, 'proc/vmstat')

    @property
    def compaction_failures_percent(self):
        if not os.path.exists(self.path):
            return 0

        fail_count = self.compact_fail
        success_count = self.compact_success
        if not success_count:
            return 0

        return int(fail_count / (success_count / 100))


class MemInfo(_BaseProcKeyValue):
    """ Interface to /proc/meminfo """
    VALID_KEYS = ['MemTotal', 'MemAvailable', 'Hugetlb', 'HugePages_Total',
                  'HugePages_Free']

    @property
    def path(self):
        return os.path.join(HotSOSConfig.data_root, 'proc/meminfo')

    @staticmethod
    def key_val_sep():
        return r':?\s+'

    @property
    def mem_total_gb(self):
        return round(self.MemTotal / (1024 * 1024))

    @property
    def mem_available_gb(self):
        return round(self.MemAvailable / (1024 * 1024))

    @property
    def hugetlb_gb(self):
        return round(self.Hugetlb / (1024 * 1024))

    @property
    def huge_pages_enabled(self):
        return self.HugePages_Total > 0

    @property
    def hugetlb_to_mem_total_percentage(self):
        return round((self.Hugetlb * 100) / self.MemTotal)

    @property
    def mem_avail_to_mem_total_percentage(self):
        return round((self.MemAvailable * 100) / self.MemTotal)

    @property
    def hugep_used_to_hugep_total_percentage(self):
        return round(100 - (self.HugePages_Free * 100) / self.HugePages_Total)


class SlabInfo():
    """ Interface to /proc/slabinfo """
    def __init__(self, filter_names=None):
        self._filter_names = filter_names or []
        self._slab_info = []
        self._load_slab_info()

    @property
    def path(self):
        return os.path.join(HotSOSConfig.data_root, "proc/slabinfo")

    @property
    def contents(self):
        return self._slab_info

    def _load_slab_info(self):
        """
        Returns a list with contents of the following columns from
        /proc/slabinfo:

            name
            num_objs
            objsize

        @param exclude_names: optional list of names to exclude.
        """
        if not os.path.exists(self.path):
            return self._slab_info

        with open(self.path, encoding='utf-8') as fd:
            for line in fd:
                exclude = False
                for name in self._filter_names:
                    if re.compile(rf'^{name}').search(line):
                        exclude = True
                        break

                if exclude:
                    continue

                sections = line.split()
                if sections[0] == '#' or sections[0] == 'slabinfo':
                    continue

                # name, num_objs, objsize
                self._slab_info.append([sections[0],
                                        int(sections[2]),
                                        int(sections[3])])

        return None

    @property
    def major_consumers(self):
        top5 = []
        top5_name = {}
        top5_num_objs = {}
        top5_objsize = {}

        # /proc/slabinfo may not exist in containers/VMs
        if not os.path.exists(self.path):
            return top5

        for line in self.contents:
            name = line[0]
            # exclude kernel memory allocations
            if name.startswith('kmalloc'):
                continue

            num_objs = line[1]
            objsize = line[2]

            for i in range(5):
                if num_objs > top5_num_objs.get(i, 0):
                    top5_num_objs[i] = num_objs
                    top5_name[i] = name
                    top5_objsize[i] = objsize
                    break

        for i in range(5):
            if top5_name.get(i):
                kbytes = top5_num_objs.get(i) * top5_objsize.get(i) / 1024
                top5.append(f"{top5_name.get(i)} ({kbytes}k)")

        return top5


class BuddyInfo():
    """ Interface to /proc/buddyinfo """
    def __init__(self):
        self._numa_nodes = []

    @property
    def path(self):
        return os.path.join(HotSOSConfig.data_root, "proc/buddyinfo")

    @property
    def nodes(self):
        """Returns list of numa nodes."""
        # /proc/buddyinfo may not exist in containers/VMs
        if not os.path.exists(self.path) or self._numa_nodes:
            return self._numa_nodes

        nodes = set()
        with open(self.path, encoding='utf-8') as fd:
            for line in fd:
                nodes.add(int(line.split()[1].strip(',')))

        self._numa_nodes = list(nodes)
        return self._numa_nodes

    def get_node_zones(self, zones_type, node):
        with open(self.path, encoding='utf-8') as fd:
            for line in fd:
                if line.split()[3] == zones_type and \
                        line.startswith(f"Node {node},"):
                    line = line.split()
                    return " ".join(line)

        return None


class MallocInfo():
    """ Helper to get memory allocation information from the system. """
    def __init__(self, node, zone):
        self.node = node
        self.zone = zone
        self._block_sizes = {}

    @property
    def block_sizes_available(self):
        if self._block_sizes:
            return self._block_sizes

        node_zones = BuddyInfo().get_node_zones(self.zone, self.node)
        if node_zones:
            # start from highest order zone (10) and work down to 0
            for order in range(10, -1, -1):
                free = int(node_zones.split()[5 + order - 1])
                self._block_sizes[order] = free

            return self._block_sizes

        return None

    @property
    def empty_order_tally(self):
        tally = 0
        for order, free in self.block_sizes_available.items():
            if not free:
                tally += order

        return tally

    @property
    def high_order_seq(self):
        """
        The number of contiguous available high-order block sizes.
        """
        available = self.block_sizes_available
        if not available:
            return 0

        # start from highest order zone (10) and work down to 0
        count = 0
        for blocks in sorted_dict(available, reverse=True).values():
            if blocks:
                break

            count += 1

        return count


class MemoryChecks():
    """ Memory checks implementation. """
    @property
    def max_unavailable_block_sizes(self):
        # 0+1+...10 is 55 so threshold is this minus the max order
        return 45

    @property
    def max_contiguous_unavailable_block_sizes(self):
        # this implies that top 5 orders are unavailable
        return 5

    @property
    def nodes_with_limited_high_order_memory_full(self):
        """
        Returns a dict of nodes and any of their zones that have limited
        high-order blocks available.
        """
        buddyinfo = BuddyInfo()
        nodes = {}
        for zone in ['Normal', 'DMA32']:
            for node in buddyinfo.nodes:
                zone_info = MallocInfo(node, zone)
                if zone_info.high_order_seq:
                    if ((zone_info.empty_order_tally >=
                            self.max_unavailable_block_sizes) or
                        (zone_info.high_order_seq >
                            self.max_contiguous_unavailable_block_sizes)):
                        availability = zone_info.block_sizes_available
                        if node in nodes:
                            nodes[node]['zones'][zone] = availability
                        else:
                            nodes[node] = {'zones': {zone: availability}}

        if nodes:
            nodes = {'nodes': nodes}

        return nodes

    @property
    def nodes_with_limited_high_order_memory(self):
        """
        Returns a list if <node>-<zone> names for zones with limited high-order
        blocks available.
        """
        nodes = []
        _nodes = self.nodes_with_limited_high_order_memory_full
        if _nodes:
            for node, zones in _nodes['nodes'].items():
                for name in zones['zones']:
                    nodes.append(f"node{node}-{name.lower()}")

        return nodes



from hotsos.core.plugins.openstack.common import (
    OpenstackEventCallbackBase,
    OpenstackEventHandlerBase,
)
from hotsos.core.search import (
    FileSearcher,
    SearchDef,
    SearchConstraintSearchSince,
)
from hotsos.core.search import CommonTimestampMatcher
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)

EXT_EVENT_META = {'network-vif-plugged': {'stages_keys':
                                          ['Preparing', 'Received',
                                           'Processing']},
                  'network-changed': {'stages_keys': ['Received',
                                                      'Refreshing']}}


class ExternalEventsCallback(OpenstackEventCallbackBase):
    """ Implements Openstack Nova external events callback. """
    event_group = 'nova.external-events'
    event_names = ['network-changed', 'network-vif-plugged']

    @staticmethod
    def _get_state_dict(event_name):
        state = {}
        for key in EXT_EVENT_META[event_name]['stages_keys']:
            state[key] = False

        return state

    def _get_output(self, event_name, results, events):
        ext_output = {}
        for event_id, event_dict in events.items():
            instance_id = event_dict['instance_id']
            data_source = event_dict['data_source']
            stages = self._get_state_dict(event_name)
            for stage in stages:
                tag = f"{instance_id}_{event_id}_{stage}"
                r = results.find_by_tag(tag, path=data_source)
                if r:
                    stages[stage] = True

            if all(values for stage, values in stages.items()):
                result = 'succeeded'
            else:
                result = 'failed'

            if result not in ext_output:
                ext_output[result] = []

            info = {'port': event_id, 'instance': instance_id}
            ext_output[result].append(info)

        return ext_output

    def _get_events_found(self, event_name, results, events):
        events_found = {}

        ext_output = self._get_output(event_name, results, events)
        if ext_output:
            for result, values in ext_output.items():
                events_found[result] = list(values)

        return events_found

    def __call__(self, event):
        events = {}

        c = SearchConstraintSearchSince(ts_matcher_cls=CommonTimestampMatcher)
        s = FileSearcher(constraint=c)
        for result in event.results:
            instance_id = result.get(1)
            event_id = result.get(2)
            result_path = event.searcher.resolve_source_id(result.source_id)
            events[event_id] = {'instance_id': instance_id,
                                'data_source': result_path}

            for stage in EXT_EVENT_META[event.name]['stages_keys']:
                expr = (r"[\d-]+ [\d:]+\.\d{3} "
                        rf".+\[instance: {instance_id}\]"
                        rf"\s+{stage}\s.*\s?event\s+{event.name}-{event_id}.?")
                tag = f"{instance_id}_{event_id}_{stage}"
                sd = SearchDef(expr, tag, hint=event.name,
                               store_result_contents=False)
                s.add(sd, result_path)

        results = s.run()

        return self._get_events_found(event.name, results, events)


class NovaExternalEventChecks(OpenstackEventHandlerBase):
    """ Implements Openstack Nova external events handler. """
    event_group = 'nova.external-events'
    summary_part_index = 1

    @summary_entry('os-server-external-events',
                   get_min_available_entry_index() + 4)
    def summary_os_server_external_events(self):
        return self.run()



from hotsos.core.plugins.sosreport import SOSReportChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class SOSReportSummary(SOSReportChecks):
    """ Implementation of SOSReport summary. """
    summary_part_index = 0

    # REMINDER: common entries are implemented in the SummaryBase base class
    #           and only application plugin specific customisations are
    #           implemented here. We use the get_min_available_entry_index() to
    #           ensure that additional entries don't clobber existing ones but
    #           conversely can also replace them by re-using their indices.

    @summary_entry('plugin-timeouts', get_min_available_entry_index())
    def summary_plugin_timeouts(self):
        if self.timed_out_plugins:
            return self.timed_out_plugins

        return None



from datetime import timedelta

from hotsos.core.host_helpers import SystemdHelper
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.requires import ServiceCheckItemsBase
from hotsos.core.ycheck.engine.properties.requires.types import (
    service_manager_common
)


class SystemdServiceCheckItems(ServiceCheckItemsBase):
    """
    Provides means to run checks on one or more systemd check item.
    """
    @property
    def _svcs_info(self):
        return SystemdHelper(self._svcs_all)


class YRequirementTypeSystemd(service_manager_common.ServiceManagerTypeBase):
    """
    Systemd requires type property. Provides support for defining checks on
    systemd resources.
    """
    override_keys = ['systemd']
    override_autoregister = True
    default_op = 'eq'

    def _check_service(self, svc, ops, started_after=None):
        """
        @param svc: SystemdService object of service we are checking.
        @param ops: list of operations to apply
        @param started_after: optional SystemdService object. If provided we
                                       will assert that svc was started after
                                       this service.
        """
        if started_after:
            a = svc.start_time
            b = started_after.start_time
            if not all([a, b]):
                log.debug("could not get start time for services %s and %s "
                          "so assuming started-after constraint is False",
                          svc.name, started_after.name)
                return False

            log.debug("%s started=%s, %s started=%s", svc.name,
                      a, started_after.name, b)
            if a < b:
                delta = b - a
            else:
                delta = a - b

            # Allow a small grace period to avoid false positives e.g.
            # on node boot when services are started all at once.
            grace = 120
            # In order for a service A to have been started after B it must
            # have been started more than GRACE seconds after B.
            if a > b:
                if delta >= timedelta(seconds=grace):
                    log.debug("svc %s started >= %ds of start of %s "
                              "(delta=%s)", svc.name, grace,
                              started_after.name, delta)
                else:
                    log.debug("svc %s started < %ds of start of %s "
                              "(delta=%s)", svc.name, grace,
                              started_after.name, delta)
                    return False
            else:
                log.debug("svc %s started before or same as %s "
                          "(delta=%s)", svc.name,
                          started_after.name, delta)
                return False

        return self.apply_ops(ops, opinput=svc.state)

    def _check_item_settings(self, svc_obj, settings, cache_info,
                             all_items):
        # pylint: disable=duplicate-code
        processes = None
        started_after_obj = None
        if isinstance(settings, str):
            state = settings
            ops = [[self.default_op, state]]
        else:
            processes = settings.get('processes')
            if processes:
                log.debug("checking systemd service processes: %s", processes)
                processes = settings.get('processes')

            op = settings.get('op', self.default_op)
            started_after = settings.get('started-after')
            started_after_obj = all_items.installed.get(started_after)
            if 'state' in settings:
                ops = [[op, settings.get('state')]]
            else:
                ops = []

        if processes and not all_items.processes_running(processes):
            log.debug("one or more systemd service processes not running "
                      "- %s", ', '.join(processes))
            return False

        cache_info[svc_obj.name]['ops'] = self.ops_to_str(ops)
        return self._check_service(svc_obj, ops,
                                   started_after=started_after_obj)

    @property
    def items_to_check(self):
        return SystemdServiceCheckItems(self.content)



import os

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import SYSCtlConfHelper
from hotsos.core.plugins.system import SystemChecks
from hotsos.core.plugintools import summary_entry


class SYSCtlChecks(SystemChecks):
    """ System sysctl checker.

    Checks system sysctl state.
    """
    summary_part_index = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_fs_sysctl = None

    @staticmethod
    def _get_values_prioritised(config, key_priorities, conffile):
        """
        Run through all key/value pairs from this config set and only keep ones
        that have not yet been supersceded by another file.

        @param config: dict if key/value pairs
        @param key_priorities: dict if found keys and the highest priority
                               attributed to them.
        @param conffile: the name of the conf file these settings are from.
                         None implies infinite priority i.e. /etc/sysctl.conf
        """
        kv_conf = {}
        for key, value in config.items():
            if conffile and key in key_priorities:
                if key_priorities[key] >= conffile:
                    continue

            key_priorities[key] = conffile
            kv_conf[key] = (value, conffile)

        return kv_conf

    def _get_charm_sysctl_d(self):
        """ Collect all key/value pairs defined under /etc/sysctl.d that were
        written by a charm.
        """
        config = {'set': {},
                  'unset': {}}

        path = os.path.join(HotSOSConfig.data_root, 'etc/sysctl.d')
        if not os.path.isdir(path):
            return config

        conffiles = ['50-ceph-osd-charm.conf',
                     '50-nova-compute.conf',
                     '50-openvswitch.conf',
                     '50-swift-storage-charm.conf',
                     '50-quantum-gateway.conf']
        key_priorities = {}
        for conffile in sorted(conffiles):
            sysctl = SYSCtlConfHelper(os.path.join(path, conffile))
            _set = self._get_values_prioritised(sysctl.setters,
                                                key_priorities,
                                                conffile)
            config['set'].update(_set)
            _unset = self._get_values_prioritised(sysctl.unsetters,
                                                  key_priorities,
                                                  conffile)
            config['unset'].update(_unset)

        return config

    def _get_sysctl_conf(self):
        config = {'set': {},
                  'unset': {}}
        path = os.path.join(HotSOSConfig.data_root, 'etc/sysctl.conf')
        if not os.path.exists(path):
            return config

        sysctl = SYSCtlConfHelper(path)
        config['set'] = self._get_values_prioritised(sysctl.setters, {}, None)
        config['unset'] = self._get_values_prioritised(sysctl.unsetters, {},
                                                       None)
        return config

    def _get_sysctl_d(self):
        """ Collect all key/value pairs defined under /etc/sysctl.d and keep
        the highest priority value for any given key.

        man sysctld specifies that the following locations are searched for
        config so we will include these:

            * /etc/sysctl.d
            * /usr/lib/sysctl.d
            * /run/sysctl.d
        """
        key_priorities = {}
        config = {'set': {},
                  'unset': {}}
        for location in ['etc', 'usr/lib', 'run']:
            path = os.path.join(HotSOSConfig.data_root, location, 'sysctl.d')
            if not os.path.isdir(path):
                continue

            for conffile in os.listdir(path):
                # Only files ending in .conf are recognised by sysctl so don't
                # count content of other files as expected.
                if not conffile.endswith('.conf'):
                    continue

                sysctl = SYSCtlConfHelper(os.path.join(path, conffile))
                _set = self._get_values_prioritised(sysctl.setters,
                                                    key_priorities,
                                                    conffile)
                config['set'].update(_set)
                _unset = self._get_values_prioritised(sysctl.unsetters,
                                                      key_priorities,
                                                      conffile)
                config['unset'].update(_unset)

        return config

    def _get_fs_sysctl(self):
        if self._cached_fs_sysctl:
            return self._cached_fs_sysctl

        sysctl = self._get_sysctl_d()
        # /etc/sysctl.conf overrides all
        _sysctl = self._get_sysctl_conf()
        sysctl['set'].update(_sysctl['set'])
        sysctl['unset'].update(_sysctl['unset'])
        self._cached_fs_sysctl = sysctl
        return self._cached_fs_sysctl

    @summary_entry('sysctl-mismatch', 10)
    def summary_sysctl_mismatch(self):
        """ Compare the values for any key set under sysctl.d and report
        an issue if any mismatches detected.
        """
        sysctl = self._get_fs_sysctl()
        mismatch = {}
        for key, config in sysctl['set'].items():
            # some keys will not be readable e.g. when inside an unprivileged
            # container so we just ignore them.
            value = config[0]
            if key in sysctl['unset']:
                s_priority = config[1]
                u_priority = sysctl['unset'][key][1]
                # None priority implies infinite or /etc/sysctl.conf
                if (s_priority is None or u_priority is None or
                        u_priority > s_priority):
                    continue

            if value != self.sysctl_all.get(key, value):
                mismatch[key] = {"actual": self.sysctl_all[key],
                                 "expected": value}

        return mismatch or None

    @summary_entry('juju-charm-sysctl-mismatch', 11)
    def summary_juju_charm_sysctl_mismatch(self):
        """ Compare the values for any key set under sysctl.d and report
        an issue if any mismatches detected.
        """
        sysctl = self._get_fs_sysctl()
        mismatch = {}
        sysctl = self._get_charm_sysctl_d()
        for key, config in sysctl['set'].items():
            value = config[0]
            if value != self.sysctl_all.get(key):
                mismatch[key] = {"conf": config[1],
                                 "actual": self.sysctl_all.get(key),
                                 "expected": value}

        return mismatch or None



# pylint: disable=too-few-public-methods

import abc


class IssueTypeBase():
    """ Base class for all implementations for issue type. """
    ISSUE_TYPE = 'issue'

    def __init__(self, msg):
        self.msg = msg

    @property
    def name(self):
        return self.__class__.__name__


class BugTypeBase(abc.ABC, IssueTypeBase):
    """ Base class for issues related to bugs. """
    ISSUE_TYPE = 'bug'

    def __init__(self, bug_id, msg):
        super().__init__(msg)
        self.id = bug_id

    @property
    @abc.abstractmethod
    def base_url(self):
        pass

    @property
    @abc.abstractmethod
    def url(self):
        pass


class CVETypeBase(BugTypeBase):
    """ Base class for issues related to CVE bugs. """
    ISSUE_TYPE = 'cve'


class HotSOSScenariosWarning(IssueTypeBase):
    """ Issue raised when one or more scenario failed to run.  """


class SystemWarning(IssueTypeBase):
    """ Issue for system level warnings. """


class KernelError(IssueTypeBase):
    """ Issue for kernel level errors. """


class KernelWarning(IssueTypeBase):
    """ Issue for kernel level warnings. """


class MemoryWarning(IssueTypeBase):
    """ Issue for memory warnings. """


class CephWarning(IssueTypeBase):
    """ Issue for ceph warnings. """


class CephHealthWarning(IssueTypeBase):
    """ Issue for Ceph cluster health warnings. """


class CephCrushWarning(IssueTypeBase):
    """ Issue for Ceph cluster CRUSH warnings. """


class CephCrushError(IssueTypeBase):
    """ Issue for Ceph cluster CRUSH errors. """


class CephOSDError(IssueTypeBase):
    """ Issue for Ceph osd errors. """


class CephOSDWarning(IssueTypeBase):
    """ Issue for Ceph osd warnings. """


class CephMonWarning(IssueTypeBase):
    """ Issue for Ceph mon warnings. """


class CephMapsWarning(IssueTypeBase):
    """ Issue for Ceph maps warnings. """


class CephMgrError(IssueTypeBase):
    """ Issue for Ceph mgr errors. """


class CephRGWWarning(IssueTypeBase):
    """ Issue for Ceph RGW warnings. """


class CephDaemonWarning(IssueTypeBase):
    """ Issue for Ceph daemon warnings. """


class CephDaemonVersionsError(IssueTypeBase):
    """ Issue for Ceph daemon versions error. """


class JujuWarning(IssueTypeBase):
    """ Issue for Juju warnings. """


class BcacheWarning(IssueTypeBase):
    """ Issue for Bcache warnings. """


class NeutronL3HAWarning(IssueTypeBase):
    """ Issue for Neutron L3HA warnings. """


class NetworkWarning(IssueTypeBase):
    """ Issue for network warnings. """


class NFSNameResolutionError(IssueTypeBase):
    """ Issue for NFS name resolution. """


class RabbitMQWarning(IssueTypeBase):
    """ Issue for RabbitMQ warnings. """


class OpenstackWarning(IssueTypeBase):
    """ Issue for Openstack warnings. """


class OVNError(IssueTypeBase):
    """ Issue for OpenvSwitch OVN errors. """


class OVNWarning(IssueTypeBase):
    """ Issue for OpenvSwitch OVN warnings. """


class OpenvSwitchWarning(IssueTypeBase):
    """ Issue for OpenvSwitch OVN errors. """


class SOSReportWarning(IssueTypeBase):
    """ Issue for SOSReport warnings. """


class SysCtlWarning(IssueTypeBase):
    """ Issue for SYSCtl warnings. """


class OpenstackError(IssueTypeBase):
    """ Issue for Openstack errors. """


class KubernetesWarning(IssueTypeBase):
    """ Issue for Kubernetes warnings. """


class PacemakerWarning(IssueTypeBase):
    """ Issue for Pacemaker warnings. """


class MySQLWarning(IssueTypeBase):
    """ Issue for MYSQL warnings. """


class LXDWarning(IssueTypeBase):
    """ Issue for LXD warnings. """


class MAASWarning(IssueTypeBase):
    """ Issue for MAAS warnings. """


class VaultWarning(IssueTypeBase):
    """ Issue for Vault warnings. """


class SSSDWarning(IssueTypeBase):
    """ Issue for SSSD warnings. """


class UbuntuCVE(CVETypeBase):
    """ Ubuntu CVE bug type """
    @property
    def base_url(self):
        return 'https://ubuntu.com/security/'

    @property
    def url(self):
        return f"{self.base_url}{self.id}"


class MitreCVE(CVETypeBase):
    """ Mitre CVE bug type """
    @property
    def base_url(self):
        return 'https://www.cve.org/CVERecord?id='

    @property
    def url(self):
        return f"{self.base_url}{self.id}"


class LaunchpadBug(BugTypeBase):
    """ Launchpad bug type """
    @property
    def base_url(self):
        return 'https://bugs.launchpad.net/bugs/'

    @property
    def url(self):
        return f"{self.base_url}{self.id}"


class Bugzilla(BugTypeBase):
    """ Bugzilla bug type """
    @property
    def base_url(self):
        return 'https://bugzilla.redhat.com/show_bug.cgi?id='

    @property
    def url(self):
        return f"{self.base_url}{self.id}"


class StoryBoardBug(BugTypeBase):
    """ Storyboard bug type """
    @property
    def base_url(self):
        return 'https://storyboard.openstack.org/#!/story/'

    @property
    def url(self):
        return f"{self.base_url}{self.id}"


class CephTrackerBug(BugTypeBase):
    """ Ceph tracker bug type """
    @property
    def base_url(self):
        return 'https://tracker.ceph.com/issues/'

    @property
    def url(self):
        return f"{self.base_url}{self.id}"



import datetime

import yaml
from hotsos.core.analytics import (
    LogEventStats,
    SearchResultIndices
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.issues import (
    IssueContext,
    IssuesManager,
    NeutronL3HAWarning,
    OpenstackWarning
)
from hotsos.core.log import log
from hotsos.core.plugins.openstack.common import (
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from hotsos.core.plugins.openstack.neutron import NeutronHAInfo
from hotsos.core import utils
from hotsos.core.utils import sorted_dict
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)

VRRP_TRANSITION_WARN_THRESHOLD = 8


class ApacheEventCallback(OpenstackEventCallbackBase):
    """ Implements Apache events callback. """
    event_group = 'apache'
    event_names = ['connection-refused']

    @staticmethod
    def _get_context_and_results(event):
        context = {}
        results = []
        for result in event.results:
            # save some context info
            path = event.searcher.resolve_source_id(result.source_id)
            if path in context:
                context[path].append(result.linenumber)
            else:
                context[path] = [result.linenumber]

            month = datetime.datetime.strptime(result.get(1), '%b').month
            day = result.get(2)
            year = result.get(3)
            addr = result.get(4)
            port = result.get(5)
            results.append({'date': f"{year}-{month}-{day}",
                            'key': f"{addr}:{port}"})

        return context, results

    def __call__(self, event):
        ports_max = {}

        context, results = self._get_context_and_results(event)

        conns_refused = self.categorise_events(event, results=results)
        for addrs in conns_refused.values():
            for addr, count in addrs.items():
                port = addr.partition(':')[2]
                # allow a small number of connection refused errors on a given
                # day
                if count < 5:
                    continue

                if port not in ports_max:
                    ports_max[port] = count
                else:
                    ports_max[port] = max(count, ports_max[port])

        if ports_max:
            msg = ('apache is reporting connection refused errors for the '
                   'following ports which could mean some services are not '
                   f"working properly - {','.join(ports_max.keys())} - please "
                   'check.')
            IssuesManager().add(OpenstackWarning(msg), IssueContext(**context))

        return sorted_dict(conns_refused)


class ApacheEventChecks(OpenstackEventHandlerBase):
    """ Implements Apache events handler. """
    summary_part_index = 8
    event_group = 'apache'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        out = self.run()
        if out:
            return {self.event_group: out}

        return None


class APIEventsCallback(OpenstackEventCallbackBase):
    """ Implements Nova API events callback. """
    event_group = 'http-requests'
    event_names = ['neutron']

    def __call__(self, event):
        results = [{'date': r.get(1),
                    'key': r.get(2)} for r in event.results]
        ret = self.categorise_events(event, results=results)
        if ret:
            return ret

        return None


class APIEvents(OpenstackEventHandlerBase):
    """ Implements Nova API events handler. """
    summary_part_index = 9
    event_group = 'http-requests'

    @summary_entry('api-info', get_min_available_entry_index() + 100)
    def summary_api_info(self):
        out = self.run()
        if out:
            return {self.event_group: out}

        return None


class AgentEventsCallback(OpenstackEventCallbackBase):
    """ Implements agent events callback. """
    event_group = 'neutron.agents'
    ovsdbapp_event_names = ['ovsdbapp-nb-leader-reconnect',
                            'ovsdbapp-sb-leader-reconnect']
    ovn_mech_driver_events = ['ovn-resource-revision-bump',
                              'ovsdb-monitor-router-binding-transitions',
                              'ovsdb-transaction-aborted']
    event_names = ['rpc-loop', 'router-spawn-events', 'router-updates']
    event_names += ovsdbapp_event_names + ovn_mech_driver_events

    @staticmethod
    def _get_event_stats(results, tag_prefix, custom_idxs=None):
        stats = LogEventStats(results, tag_prefix, custom_idxs=custom_idxs)
        stats.run()
        top5 = stats.get_top_n_events_sorted(5)
        if not top5:
            return None

        return {"top": top5,
                "stats": stats.get_event_stats()}

    def _ovsdb_monitor_router_binding_transitions_event(self, event):
        """
        The ovsdbmonitor will print bindings which will eventually
        reflect transitions so we need to filter out contiguous bindings
        to the same chassis so that we are just left with transitions
        that we can then count.

        :param event: EventCheckResult
        """
        agent = event.section_name
        results = []
        for r in event.results:
            results.append({'date': r.get(1), 'time': r.get(2),
                            'router': r.get(3), 'key': r.get(4)})

        results = sorted(results,
                         key=lambda e:
                         datetime.datetime.strptime(
                             f"{e['date']} {e['time']}",
                             '%Y-%m-%d %H:%M:%S'))

        new_results = []
        last_known = {}
        host_key = 'key'
        for item in results:
            router = item['router']
            if router not in last_known:
                last_known[router] = item
            elif last_known[router][host_key] != item[host_key]:
                # remember hosts port transitioned from
                new_results.append(last_known[router])
                last_known[router] = item

        ret = self.categorise_events(event, results=new_results)
        if ret:
            return {event.name: ret}, agent

        return None

    def _ml2_ovs_agent_events(self, event):
        """
        :param event: EventCheckResult
        """
        agent = event.section_name
        ret = self._get_event_stats(event.results, event.search_tag)
        if ret:
            return {event.name: ret}, agent

        return None

    def __call__(self, event):
        agent = event.section_name
        if event.name == 'ovsdb-monitor-router-binding-transitions':
            return self._ovsdb_monitor_router_binding_transitions_event(event)

        if event.name in ['rpc-loop', 'router-spawn-events']:
            return self._ml2_ovs_agent_events(event)

        if event.name in ['router-updates']:
            sri = SearchResultIndices(
                event_id=4, metadata=3, metadata_key="router"
            )
            ret = self._get_event_stats(event.results, event.search_tag,
                                        custom_idxs=sri)
            if ret:
                return {event.name: ret}, agent

            return None

        # All other events should be covered by the following
        if event.name == 'ovn-resource-revision-bump':
            ret = self.categorise_events(
                event,
                options=self.EventProcessingOptions(max_results_per_date=5)
            )
        elif event.name == 'ovsdb-transaction-aborted':
            ret = {}
            for result in event.results:
                _date = result.get(1)
                if _date in ret:
                    ret[_date] += 1
                else:
                    ret[_date] = 1

            ret = sorted_dict(ret)
        else:
            ret = self.categorise_events(event)

        if ret:
            return {event.name: ret}, agent

        return None


class NeutronAgentEventChecks(OpenstackEventHandlerBase):
    """
    Loads events we want to check from definitions yaml and executes them. The
    results are sorted by date and the "top 5" are presented along with stats
    on the full set of samples.
    """
    summary_part_index = 7
    event_group = 'neutron.agents'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        # NOTE: order is important here
        agents = ['neutron-server', 'neutron-l3-agent', 'neutron-ovs-agent']
        out = self.run() or {}
        return {agent: out[agent] for agent in agents if agent in out}


class OctaviaAgentEventsCallback(OpenstackEventCallbackBase):
    """ Implements OpenStack octavia agent events callback. """
    event_group = 'octavia'
    event_names = ['lb-failover-auto', 'lb-failover-manual',
                   'amp-missed-heartbeats']

    def __call__(self, event):
        if event.name in ['lb-failover-auto', 'lb-failover-manual']:
            results = []
            for e in event.results:
                payload = yaml.safe_load(e.get(3))
                lb_id = payload.get('load_balancer_id')
                if lb_id is None:
                    continue

                results.append({'date': e.get(1), 'time': e.get(2),
                                'key': lb_id})

            ret = self.categorise_events(
                event,
                results=results,
                options=self.EventProcessingOptions(key_by_date=False)
            )
            if ret:
                failover_type = event.name.rpartition('-')[2]
                return {failover_type: ret}, 'lb-failovers'
        else:
            missed_heartbeats = self.categorise_events(
                event,
                options=self.EventProcessingOptions(key_by_date=False)
            )
            if not missed_heartbeats:
                return None

            # sort each amp by occurences
            for ts_date, amps in missed_heartbeats.items():
                amps_sorted = utils.sorted_dict(amps, key=lambda e: e[1],
                                                reverse=True)
                missed_heartbeats[ts_date] = amps_sorted

            # then sort by date
            return utils.sorted_dict(missed_heartbeats)

        return None


class OctaviaAgentEventChecks(OpenstackEventHandlerBase):
    """ Implements Octavia agent events handler. """
    summary_part_index = 10
    event_group = 'octavia'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        out = self.run()
        if out:
            return {self.event_group: out}

        return None


class PCINotFoundCallback(OpenstackEventCallbackBase):
    """ Implements Nova PCINotFound error events callback. """
    event_group = 'nova.nova-compute'
    event_names = ['pci-dev-not-found']

    def __call__(self, event):
        ret = self.categorise_events(event)
        if ret:
            return ret, 'PciDeviceNotFoundById'

        return None


class NovaComputeEventChecks(OpenstackEventHandlerBase):
    """ Implements Nova Compute events handler. """
    summary_part_index = 11
    event_group = 'nova.nova-compute'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        out = self.run()
        if out:
            return {'nova': out}

        return None


class ApparmorCallback(OpenstackEventCallbackBase):
    """ Implements Apparmor events callback. """
    event_group = 'apparmor'
    event_names = ['nova', 'neutron']

    def __call__(self, event):
        # pylint: disable=duplicate-code
        results = [{'date': f"{r.get(1)} {r.get(2)}",
                    'time': r.get(3),
                    'key': r.get(4)} for r in event.results]
        ret = self.categorise_events(
            event,
            results=results,
            options=self.EventProcessingOptions(key_by_date=False),
        )
        if ret:
            # event.name must be the service name, event.section is the aa
            # action.
            return {event.name: ret}, event.section_name

        return None


class AgentApparmorChecks(OpenstackEventHandlerBase):
    """ Implements Apparmor events handler. """
    summary_part_index = 12
    event_group = 'apparmor'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        out = self.run()
        if out:
            return {self.event_group: out}

        return None


class L3HACallback(OpenstackEventCallbackBase):
    """ Implements Neutron L3HA events callback. """
    event_group = 'neutron.ml2-routers'
    event_names = ['vrrp-transitions']

    @staticmethod
    def check_vrrp_transitions(transitions):
        # there will likely be a large number of transitions if we look across
        # all time so dont run this check.
        if HotSOSConfig.use_all_logs:
            return None

        max_transitions = 0
        warn_count = 0
        threshold = VRRP_TRANSITION_WARN_THRESHOLD
        for router in transitions:
            r = transitions[router]
            _transitions = sum(t for d, t in r.items())
            if _transitions > threshold:
                max_transitions = max(_transitions, max_transitions)
                warn_count += 1

        if warn_count:
            msg = (f"{warn_count} router(s) have had more than {threshold} "
                   f"vrrp transitions (max={max_transitions}) in the last 24 "
                   "hours.")
            IssuesManager().add(NeutronL3HAWarning(msg))

        return None

    def __call__(self, event):
        results = []
        for r in event.results:
            router = NeutronHAInfo().find_router_with_vr_id(r.get(3))
            if not router:
                log.debug("could not find router with vr_id %s", r.get(3))
                continue

            results.append({'date': r.get(1), 'time': r.get(2),
                            'key': router.uuid})

        transitions = self.categorise_events(
            event,
            results=results,
            options=self.EventProcessingOptions(key_by_date=False),
        )
        if transitions:
            # run checks
            self.check_vrrp_transitions(transitions)
            # add info to summary
            return {'transitions': transitions}, 'keepalived'

        return None


class NeutronL3HAEventCheckJournalCtl():
    """ Implements Neutron L3HA event check journalctl. """
    @staticmethod
    def args():
        """ Args callback for event cli command """
        args = []
        kwargs = {'unit': 'neutron-l3-agent'}
        if not HotSOSConfig.use_all_logs:
            kwargs['date'] = CLIHelper().date(format="--iso-8601")

        return args, kwargs


class NeutronL3HAEventChecks(OpenstackEventHandlerBase):
    """ Implements Neutron L3HA events handler. """
    summary_part_index = 13
    event_group = 'neutron.ml2-routers'

    @summary_entry('agent-checks', get_min_available_entry_index() + 101)
    def summary_agent_checks(self):
        out = self.run()
        if out:
            return {'neutron-l3ha': out}

        return None



import abc
import contextlib
from collections import UserDict

from hotsos.core.log import log
from hotsos.core.search import (
    FileSearcher,
    SearchConstraintSearchSince,
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.search import CommonTimestampMatcher
from hotsos.core.exceptions import NameNotSetError


class SearchRegistryError(Exception):
    """ Generic search registry exception. """
    def __init__(self, msg):
        self.msg = msg


class SearchRegistryKeyConflict(Exception):
    """ Search registry key conflict exception.

    This is raised when an attempt is made to register a new item in the
    registry using a key for which an item has already been registered.
    """
    def __init__(self, key, all_keys):
        self.key = key
        self.all_keys = all_keys

    def __str__(self):
        return (f"'{self.key}' key already exists in search registry. "  # noqa, pylint: disable=consider-using-f-string
                "Available keys are:\n      - {}".
                format('\n      - '.join(self.all_keys)))


class SearchRegistryKeyNotFound(Exception):
    """ Search registry key not found exception.

    This is raised when an attempt is made to retrieve an item from the
    registry using a key for which an item does not exist.
    """
    def __init__(self, key, all_keys):
        self.key = key
        self.all_keys = all_keys

    def __str__(self):
        return (f"'{self.key}' not found in search registry. Available keys "  # noqa, pylint: disable=consider-using-f-string
                "are:\n      - {}".
                format('\n      - '.join(self.all_keys)))


SEARCHES_TO_BE_REGISTERED = []


class GlobalSearcher(contextlib.AbstractContextManager, UserDict):
    """
    A shared searcher used to load searches from as many sources as
    possible prior to execution so as to minimise the amount of times we have
    to walk the same files. The is particularly useful for YAML checks
    i.e. events and scenarios that contain many search properties.

    Search properties are registered using their resolve path i.e. to the
    parent events or scenarios check in the yaml tree that contains them.
    Entries must be unique. Once all searches are loaded they are executed and
    their results are made available to anyone who wants them. Results are
    accessed using the yaml dot path used to register the search.
    """

    def __init__(self):
        constraint = SearchConstraintSearchSince(
                                         ts_matcher_cls=CommonTimestampMatcher)
        self._loaded_searches = []
        self._results = None
        self._searcher = FileSearcher(constraint=constraint)
        log.debug("creating new global searcher (%s)", self._searcher)
        super().__init__()

    def __exit__(self, *exc_details):
        pass

    def __setitem__(self, key, item):
        """ Register a new search.

        @param key: search property YAML resolve path.
        @param item: search property object.
        """
        if key in self:
            raise SearchRegistryKeyConflict(key, list(self.data))

        log.debug("adding key=%s to search registry", key)
        super().__setitem__(key, item)

    def __getitem__(self, key):
        """ Access a search entry.

        @param key: search property YAML resolve path.
        """
        try:
            return super().__getitem__(key)
        except KeyError:
            raise SearchRegistryKeyNotFound(key, list(self.data)) from KeyError

    def set_loaded(self, label):
        """ After loading a set of searches it is useful to set a label as an
        easy to later check if they have been loaded since this can only be
        done once per label.

        @param label: string label to mark a set of searches as registered.
        """
        if label in self._loaded_searches:
            raise SearchRegistryError("Search Registry has already been "
                                      f"loaded by label={label}")

        self._loaded_searches.append(label)

    def is_loaded(self, label):
        return label in self._loaded_searches

    @property
    def searcher(self):
        return self._searcher

    @property
    def results(self):
        """
        Execute searches if first time called and use cached results for
        future calls.
        """
        if self._results is not None:
            log.debug("using cached global searcher results")
            return self._results

        for s in SEARCHES_TO_BE_REGISTERED:
            if not s().validate(HotSOSConfig.plugin_name):
                continue

            self.add_search(s.simple_search(), s.sequence_search(),
                            s.passthrough_results, s.paths())

        log.debug("fetching global searcher results")
        self._results = self.searcher.run()
        return self._results

    def add_search(self, simple_search, sequence_search, passthrough_results,
                   paths):
        """
        Register a new search in the registry and add it to the global
        searcher.

        @param simple_search: SearchDef object.
        @param sequence_search: SequenceSearchDef object.
        @param passthrough_results: True or False.
        @param paths: list of one or more paths.
        """
        if simple_search is not None:
            searchdef = simple_search
        else:
            searchdef = sequence_search

        tag = searchdef.tag
        self[tag] = {'search_tag': tag,
                     'passthrough_results': passthrough_results,
                     'simple_search': simple_search,
                     'sequence_search': sequence_search}

        for path in paths:
            log.debug("loading search (tag=%s, input_path=%s)", tag,
                      path)
            self.searcher.add(searchdef, path,
                              allow_global_constraints=True)

    def run(self):
        _ = self.results


class GlobalSearcherPreloaderBase():
    """ Used by plugin components to preload the global searcher. """

    @staticmethod
    def _find_search_prop_parent(items, path):
        """
        Walk down path until we hit the item containing the property. We skip
        root/plugin name at start and ".<property>" at the end.

        @param items: YDefsSection object representing the entire tree of
                      properties.
        @param path: property resolve path.
        """
        item = None
        for branch in path.split('.')[1:-1]:
            item = getattr(items if item is None else item, branch)

        return item

    @staticmethod
    def _load_item_search(global_searcher, search_property, search_input):
        """ Load search property information into searcher.

        @param global_searcher: GlobalSearcher object
        @param search_property: YPropertySearch object
        @param search_input: YPropertyInput object
        """
        if len(search_input.paths) == 0:
            return

        allow_constraints = True
        if search_input.command:
            # don't apply constraints to command outputs
            allow_constraints = False

        # Add to registry in case it is needed by handlers e.g. for
        # sequence lookups.
        # IMPORTANT: do not store reference to ycheck property itself.
        tag = search_property.unique_search_tag
        global_searcher[tag] = {'search_tag': tag,
                                'passthrough_results':
                                    search_property.passthrough_results,
                                'simple_search':
                                    search_property.simple_search,
                                'sequence_search':
                                    search_property.sequence_search}

        for path in search_input.paths:
            log.debug("loading search (tag=%s, input_path=%s, "
                      "resolve_path=%s)", tag, path,
                      search_property.override_path)
            search_property.load_searcher(global_searcher.searcher, path,
                                          allow_constraints=allow_constraints)

    @staticmethod
    @abc.abstractmethod
    def skip_filtered(path):
        pass

    @abc.abstractmethod
    def preload_searches(self, global_searcher):
        pass


class GlobalSearcherAutoRegisterMeta(type):
    """ Add to list of searches to be registered at runtime. """

    def __init__(cls, _name, _mro, _members):
        if cls.unique_search_tag is None:
            return

        SEARCHES_TO_BE_REGISTERED.append(cls)


class GlobalSearcherAutoRegisterBase(metaclass=GlobalSearcherAutoRegisterMeta):
    """
    Generic interface for loading search definitions into the global searcher.
    The attributes of this class are intentionally similar to those of
    hotsos.core.ycheck.engine.properties.search so as to be able to give them
    equivalent meaning but this is really intended for use by code that wants
    to perform searches but that is not using YPropertySearch (i.e. not event
    or scenario yaml).
    """
    # This must be set to the same value as used by corresponding
    # implementation of PluginPartBase.plugin_name.
    plugin_name = None
    # This will be used to register and lookup the search in the registry and
    # must not conflict with any existing tags.
    unique_search_tag = None

    def validate(self, plugin_name):
        """
        This is used to ensure that we only load searches applicable to the
        plugin currently being executed.

        Returns True if we are running the context of the plugin to which this
        search belongs.
        """
        if self.plugin_name is None:
            raise NameNotSetError(
                f"{self.__class__.__name__}.plugin_name must be set "
                "to the name of the plugin in which it will run "
                f"(currently={self.plugin_name})")

        return plugin_name == self.plugin_name

    @property
    def passthrough_results(self):
        return False

    @classmethod
    def simple_search(cls):
        """ Returns SearchDef object.

        NOTE: either this or sequence_search must be implemented but not both.
        """

    @classmethod
    def sequence_search(cls):
        """ Returns a SequenceSearchDef object.

        NOTE: either this or simple_search must be implemented but not both.
        """

    @classmethod
    @abc.abstractmethod
    def paths(cls):
        """ Returns a list of one or more paths to search. """
